import logging

logger = logging.getLogger(__name__)


def reboot(isvgAppliance, check_mode=False, force=False):
    """
    Reboot the appliance
    """
    if check_mode is True:
        return isvgAppliance.create_return_object(changed=True)
    else:
        # obtain appliance lastboot time
        import ibmsecurity.isvg.firmware
        import time
        ret_obj_appliance = ibmsecurity.isvg.firmware.get(isvgAppliance)
        for firm in ret_obj_appliance['data']:
            if firm['active'] is True:
                before_reboot_last_boot = firm['last_boot']
                logger.info(
                    "Active partition last boot time {0} before reboot process initiated.".format(before_reboot_last_boot))

        ret_obj = isvgAppliance.invoke_post("Rebooting appliance",
                                         "/diagnostics/restart_shutdown/reboot",
                                         {})

        # Depending on resource allocated, isvg appliance can take up to 30s or more before it reboots
        # after it is being told to do so.
        # Loop until appliance returns an error, or reboot is detected, before returning control.
        try:
            while True:
                ret_obj_appliance = ibmsecurity.isvg.firmware.get(isvgAppliance)
                for firm in ret_obj_appliance['data']:
                    if firm['active'] is True:
                        last_boot = firm['last_boot']
                        logger.info(
                            "Active partition last boot time {0} after reboot process initiated.".format(last_boot))
                        if last_boot > before_reboot_last_boot:
                            raise Exception ("Break out of loop")
                time.sleep(15)
        except Exception as e:
            logger.debug("Exception occured: {0}. Assuming appliance has now initiated reboot process".format(e))
            pass

        return ret_obj


def shutdown(isvgAppliance, check_mode=False, force=False):
    """
    Shutdown the appliance
    """
    if check_mode is True:
        return isvgAppliance.create_return_object(changed=True)
    else:
        return isvgAppliance.invoke_post("Shutting down appliance",
                                         "/diagnostics/restart_shutdown/shutdown",
                                         {})


def _changes_available(isvgAppliance):
    """
    Check for pending changes on the appliance
    :param isvgAppliance:
    :return:
    """
    #    changes = isvgAppliance.invoke_get("Get pending changes",
    #                                       "/pending_changes")
    #    logger.debug("Pending changes on appliance are:")
    #    logger.debug(changes['data'])
    #
    #    change_count = isvgAppliance.invoke_get("Get pending changes count",
    #                                            "/pending_changes/count")
    #    logger.debug("Pending change count on appliance is:")
    #    logger.debug(change_count['data'])
    #
    #    if change_count['data']['count'] > 0 or len(changes['data']['changes']) > 0:
    #        logger.info("pending changes found")
    #        return True
    #    else:
    #        return False

    return True


def commit(isvgAppliance, check_mode=False, force=False):
    """
    Commit the current pending changes.
    """
    if force is True or _changes_available(isvgAppliance) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_get("Committing the changes",
                                            "/pending_changes/deploy",
                                            {})

    return isvgAppliance.create_return_object()


def commit_and_restart(isvgAppliance, check_mode=False, force=False):
    """
    Commit and Restart LMI
    :param isvgAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _changes_available(isvgAppliance) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post("Commit and Restart",
                                             "/restarts/commit_and_restart",
                                             {})
    return isvgAppliance.create_return_object()


def rollback(isvgAppliance, check_mode=False, force=False):
    """
    Rollback the current pending changes.
    """
    if force is True or _changes_available(isvgAppliance) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_get("Rollback the changes",
                                            "/pending_changes/forget")

    return isvgAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

def get_network(isamAppliance, application_interface, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Application Interface Statistics
    """
    return isamAppliance.invoke_get("Retrieving the Application Interface Statistics",
                                    f"/analysis/interface_statistics.json{tools.create_query_string(prefix=application_interface, timespan=statistics_duration)}",requires_model=requires_model)


def get_rp_junction(isamAppliance, instance, date, duration, check_mode=False, force=False):
    """
    Retrieving junction average response times for a Reverse Proxy instance
    """
    return isamAppliance.invoke_get("Retrieving junction average response times for a Reverse Proxy instance",
                                    f"/analysis/reverse_proxy_traffic/reqtime{tools.create_query_string(date=date, duration=duration, instance=instance)}",requires_model=requires_model)


def get_rp_health_summary(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a summary of Reverse Proxy health
    """
    return isamAppliance.invoke_get("Retrieving a summary of Reverse Proxy health",
                                    "/wga/widgets/health.json",requires_model=requires_model)


def get_rp_throughput_summary(isamAppliance, date, duration, aspect, summary=None, check_mode=False, force=False):
    """
    Retrieving a summary of throughput for all Reverse Proxy instances
    """
    headers = {'Accept': 'application/json', 'range': 'items=0-24'}
    return isamAppliance.invoke_get_with_headers("Retrieving a summary of throughput for all Reverse Proxy instances",
                                                 f"/analysis/reverse_proxy_traffic/throughput/{tools.create_query_string(summary=summary, date=date, duration=duration, aspect=aspect)}",
                                                                               requires_model=requires_model,
                                                                               headers=headers)


def get_rp_throughput(isamAppliance, instance, date, duration, check_mode=False, force=False):
    """
    Retrieving throughput records for a specific Reverse Proxy instance
    """
    return isamAppliance.invoke_get("Retrieving throughput records for a specific Reverse Proxy instance",
                                    f"/analysis/reverse_proxy_traffic/throughput/{instance}{tools.create_query_string(date=date, duration=duration)}",requires_model=requires_model)


def get_rp_traffic_summary(isamAppliance, instance, date, duration, aspect, summary=True, check_mode=False,
                           force=False):
    """
    Retrieving a summary of traffic by Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving a summary of traffic by Junction or User-Agent on a Reverse Proxy instance",
        f"/analysis/reverse_proxy_traffic/traffic/instance/{instance}/{tools.create_query_string(summary=summary, date=date, duration=duration, aspect=aspect)}",requires_model=requires_model)


def get_rp_traffic(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False, force=False):
    """
    Retrieving a summary of traffic records for a specific Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving a summary of traffic records for a specific Junction or User-Agent on a Reverse Proxy instance",
        f"/analysis/reverse_proxy_traffic/traffic/instance/{instance}/{aspect}/{aspect_identifier}{tools.create_query_string(date=date, duration=duration)}",requires_model=requires_model)


def get_rp_traffic_detail(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False,
                          force=False):
    """
    Retrieving detailed traffic records for a specific Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving detailed traffic records for a specific Junction or User-Agent on a Reverse Proxy instance",
        f"/analysis/reverse_proxy_traffic/traffic/instance/{instance}/{aspect}/{aspect_identifier}/{tools.create_query_string(date=date, duration=duration, aspect=aspect)}",requires_model=requires_model)


def get_rp_traffic_detail_aspect(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False,
                                 force=False):
    """
    Retrieving detailed traffic records for a specific User-Agent on a specific junction in a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving detailed traffic records for a specific User-Agent on a specific junction in a Reverse Proxy instance",
        f"/analysis/reverse_proxy_traffic/traffic/instance/{instance}/{aspect}/{aspect_identifier}/{aspect}/{aspect_identifier}{tools.create_query_string(date=date, duration=duration)}",requires_model=requires_model)


def get_rp_waf_events(isamAppliance, instance, date, duration, type, check_mode=False,
                      force=False):
    """
    Retrieving security action events for a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving security action events for a Reverse Proxy instance",
        f"/analysis/reverse_proxy_traffic/pam_events{tools.create_query_string(date=date, duration=duration, instance=instance, type=type)}",requires_model=requires_model)


def get_cpu(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the CPU Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the CPU Usage Statistics",
        f"/statistics/systems/cpu.json{tools.create_query_string(timespan=statistics_duration)}",requires_model=requires_model)


def get_memory(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the Memory Usage Statistics",
        f"/statistics/systems/memory.json{tools.create_query_string(timespan=statistics_duration)}",requires_model=requires_model)


def get_storage(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the Storage Usage Statistics",
        f"/statistics/systems/storage.json{tools.create_query_string(timespan=statistics_duration)}",requires_model=requires_model)



import logging
from ibmsecurity.utilities import tools
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/access-policies/"
requires_modules = ["federation", "mga"]
requires_version = "9.0.4.0"


def get_all(isamAppliance, sortBy=None, count=None, start=None, filter=None, check_mode=False, force=False):
    """
    Retrieve a list of access policies
    """
    return isamAppliance.invoke_get("Retrieve a list of access policies", "{0}{1}".format(uri,
                                                                                          tools.create_query_string(
                                                                                              sortBy=sortBy,
                                                                                              count=count, start=start,
                                                                                              filter=filter)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific access policy
    """
    ret_obj = search(isamAppliance, name, check_mode, force)

    warnings = []
    if ret_obj['data'] == {}:
        return isamAppliance.create_return_object(warnings=[f"Access policy with name {name} not found."])
    else:
        return _get(isamAppliance, ret_obj['data'])


def _get(isamAppliance, access_policy_id):
    """
    Internal function to get data using "id" - used to avoid extra calls
    """
    return isamAppliance.invoke_get("Retrieve a specific access policy", f"{uri}{access_policy_id}",
                                    requires_modules=requires_modules, requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search access policy by name
    """
    ret_obj = get_all(isamAppliance, check_mode=check_mode, force=force)

    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            return_obj['data'] = obj['id']
            logger.debug(f"Found id: {obj['id']}")
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, file=None, content=None, type='JavaScript', category='OIDC', check_mode=False,
        force=False):
    """
    Creating or Modifying an Access Policy
    """
    if content is None or content == '':
        if file is None or file == '':
            return isamAppliance.create_return_object(
                warnings=["Need to pass content or file for set() to work."])
        else:
            with open(file, 'r') as contentFile:
                content = contentFile.read()
    ret_obj = search(isamAppliance, name=name)
    if ret_obj['data'] == {}:
        # Force the add - we already know connection does not exist
        return add(isamAppliance, name=name, content=content, category=category, type=type,
                   check_mode=check_mode, force=True)
    else:
        # Update request
        return update(isamAppliance, name=name, content=content, check_mode=check_mode, force=force)


def add(isamAppliance, name, content, type="JavaScript", category="OIDC", check_mode=False, force=False):
    """
    Create a new access policy
    """
    ret_obj = search(isamAppliance, name)
    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new access policy", uri,
                {
                    "name": name,
                    "type": type,
                    "content": content,
                    "category": category
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an access policy
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an access policy",
                f"{uri}{ret_obj['data']}", requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, content, check_mode=False, force=False):
    """
    Update a specified access policy
    """
    update_required = False
    ret_obj = search(isamAppliance, name)
    id = ret_obj['data']
    if force is False:
        if id:
            ret_obj_content = _get(isamAppliance, id)
            # Having to strip whitespace to get a good comparison (suspect carriage returns added after save happens)
            if (ret_obj_content['data']['properties']['content']).strip() != (content).strip():
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified access policy",
                f"{uri}{id}",
                {
                    'content': content
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def export_file(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Export a specific access policy
    """
    import os.path
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or (ret_obj['data'] != {} and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            id = ret_obj['data']
            return isamAppliance.invoke_get_file(
                "Export a specific access policy",
                f"{uri}{id}/file",
                filename, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def upload(isamAppliance, name, file, type='JavaScript', category="OIDC", check_mode=False, force=False):
    """
    Import a new access policy
    """
    ret_obj = search(isamAppliance, name)
    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Import a new access policy", uri,
                [
                    {
                        'file_formfield': 'file',
                        'filename': file,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    "name": name,
                    "type": type,
                    "category": category
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def import_file(isamAppliance, name, file, check_mode=False, force=False):
    """
    Import a access policy
    """
    update_required = False
    if force is False:
        ret_obj = search(isamAppliance, name)
        if ret_obj['data'] != {}:
            ret_obj_content = _get(isamAppliance, ret_obj['data'])
            with open(file, 'r') as infile:
                content = infile.read()
            if (ret_obj_content['data']['properties']['content']).strip() != content.strip():
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Import a access policy",
                f"{uri}{ret_obj['data']}/file",
                [
                    {
                        'file_formfield': 'file',
                        'filename': file,
                        'mimetype': 'application/file'
                    }
                ],
                {}, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare access policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        obj = _get(isamAppliance1, obj['id'])['data']
        del obj['id']
    for obj in ret_obj2['data']:
        obj = _get(isamAppliance2, obj['id'])['data']
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
from ibmsecurity.isam.aac.authentication.rsa_otp.all import get

logger = logging.getLogger(__name__)

uri = "/iam/access/v8/otp/config/rsa"
requires_modules = ["mga"]
requires_version = "8.0.0.0"


def import_file(isamAppliance, node, filename, check_mode=False, force=False):
    """
    Import securid file for a node

    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_post_files(
            "Import securid file for a node",
            f"{uri}/securid/{node}",
            [
                {
                    'file_formfield': 'file',
                    'filename': filename,
                    'mimetype': 'application/file'
                }
            ],
            {},
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()


def delete(isamAppliance, node, check_mode=False, force=False):
    """
    Delete the securid file from a node

    """

    ret_obj = get(isamAppliance)
    delete_required = False

    for obj in ret_obj['data']:
        if obj['fileName'] == "securid":
            if 'importTimestamp' in obj:
                delete_required = True

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_delete(
                "Delete the securid file from a node",
                f"{uri}/securid/{node}",
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/openid"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve all openid entries
    """
    return isvgAppliance.invoke_get("Retrieve all openid entries", "{0}".format(uri))


def search(isvgAppliance, name, check_mode=False, force=False):
    """
    Search for existing openid.
    Just care for presence of openid, not its actual value.
    """
    ret_obj = get_all(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found openid entry {0}".format(name))
            return_obj['data'] = obj
            return_obj['rc'] = 0
            break

    return return_obj


def add(isvgAppliance, name, clientID, secret, certAlias, domains, issuerIdentifier, redirectToRPHostPort, userRealm, authUrl, tokenUrl, jwkUrl, logoffUrl, discoveryUrl, advanced, signAlgorithm, userIDToCreateSubject, scope, check_mode=False, force=False):
    """
    Create an openid entry
    """
    if force is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            # Create a simple json with just the openid entry
            openid_json = {
                "name": name,
                "clientID": clientID,
                "secret": secret,
                "certAlias": certAlias,
                "domains": domains,
                "issuerIdentifier": issuerIdentifier,
                "redirectToRPHostPort": redirectToRPHostPort,
                "userRealm": userRealm,
                "authUrl": authUrl,
                "tokenUrl": tokenUrl,
                "jwkUrl": jwkUrl,
                "logoffUrl": logoffUrl,
                "discoveryUrl": discoveryUrl,
                "advanced": advanced,
                "signAlgorithm": signAlgorithm,
                "scope": scope,
                "userIDToCreateSubject": userIDToCreateSubject
            }

            return isvgAppliance.invoke_post(
                "Create a new openid entry", uri, openid_json)

    return isvgAppliance.create_return_object(changed=False)


def delete(isvgAppliance, name, check_mode=False, force=False):
    """
    Delete an openid entry
    """
    ret_obj = search(isvgAppliance, name, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            uuid = ret_obj['data']['uuid']
            return isvgAppliance.invoke_delete(
                "Delete a openid entry", "{0}/{1}".format(uri, uuid), warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def update(isvgAppliance, name, clientID, secret, certAlias, domains, issuerIdentifier, redirectToRPHostPort, userRealm, authUrl, tokenUrl, jwkUrl, logoffUrl, discoveryUrl, advanced, signAlgorithm, userIDToCreateSubject, scope, check_mode=False, force=False):
    """
    Updating an openid entry
    """
    ret_obj = search(isvgAppliance, name, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    uuid = ret_obj['data']['uuid']

    needs_update = False

    # Create a simple json with just the attributes
    json_data = {
        "name": name,
        "uuid": uuid
    }

    if clientID is not None:
        json_data['clientID'] = clientID
    elif 'clientID' in ret_obj['data']:
        if ret_obj['data']['clientID'] is not None:
            json_data['clientID'] = ret_obj['data']['clientID']
        else:
            del ret_obj['data']['clientID']
    if secret is not None:
        json_data['secret'] = secret
    elif 'secret' in ret_obj['data']:
        if ret_obj['data']['secret'] is not None:
            json_data['secret'] = ret_obj['data']['secret']
        else:
            del ret_obj['data']['secret']
    if certAlias is not None:
        json_data['certAlias'] = certAlias
    elif 'certAlias' in ret_obj['data']:
        if ret_obj['data']['certAlias'] is not None:
            json_data['certAlias'] = ret_obj['data']['certAlias']
        else:
            del ret_obj['data']['certAlias']
    if domains is not None:
        json_data['domains'] = domains
    elif 'domains' in ret_obj['data']:
        if ret_obj['data']['domains'] is not None:
            json_data['domains'] = ret_obj['data']['domains']
        else:
            del ret_obj['data']['domains']
    if issuerIdentifier is not None:
        json_data['issuerIdentifier'] = issuerIdentifier
    elif 'issuerIdentifier' in ret_obj['data']:
        if ret_obj['data']['issuerIdentifier'] is not None:
            json_data['issuerIdentifier'] = ret_obj['data']['issuerIdentifier']
        else:
            del ret_obj['data']['issuerIdentifier']
    if redirectToRPHostPort is not None:
        json_data['redirectToRPHostPort'] = redirectToRPHostPort
    elif 'redirectToRPHostPort' in ret_obj['data']:
        if ret_obj['data']['redirectToRPHostPort'] is not None:
            json_data['redirectToRPHostPort'] = ret_obj['data']['redirectToRPHostPort']
        else:
            del ret_obj['data']['redirectToRPHostPort']
    if userRealm is not None:
        json_data['userRealm'] = userRealm
    elif 'userRealm' in ret_obj['data']:
        if ret_obj['data']['userRealm'] is not None:
            json_data['userRealm'] = ret_obj['data']['userRealm']
        else:
            del ret_obj['data']['userRealm']
    if authUrl is not None:
        json_data['authUrl'] = authUrl
    elif 'authUrl' in ret_obj['data']:
        if ret_obj['data']['authUrl'] is not None:
            json_data['authUrl'] = ret_obj['data']['authUrl']
        else:
            del ret_obj['data']['authUrl']
    if tokenUrl is not None:
        json_data['tokenUrl'] = tokenUrl
    elif 'tokenUrl' in ret_obj['data']:
        if ret_obj['data']['tokenUrl'] is not None:
            json_data['tokenUrl'] = ret_obj['data']['tokenUrl']
        else:
            del ret_obj['data']['tokenUrl']
    if logoffUrl is not None:
        json_data['logoffUrl'] = logoffUrl
    elif 'logoffUrl' in ret_obj['data']:
        if ret_obj['data']['logoffUrl'] is not None:
            json_data['logoffUrl'] = ret_obj['data']['logoffUrl']
        else:
            del ret_obj['data']['logoffUrl']
    if jwkUrl is not None:
        json_data['jwkUrl'] = jwkUrl
    elif 'jwkUrl' in ret_obj['data']:
        if ret_obj['data']['jwkUrl'] is not None:
            json_data['jwkUrl'] = ret_obj['data']['jwkUrl']
        else:
            del ret_obj['data']['jwkUrl']
    if discoveryUrl is not None:
        json_data['discoveryUrl'] = discoveryUrl
    elif 'discoveryUrl' in ret_obj['data']:
        if ret_obj['data']['discoveryUrl'] is not None:
            json_data['discoveryUrl'] = ret_obj['data']['discoveryUrl']
        else:
            del ret_obj['data']['discoveryUrl']
    if advanced is not None:
        json_data['advanced'] = advanced
    elif 'advanced' in ret_obj['data']:
        if ret_obj['data']['advanced'] is not None:
            json_data['advanced'] = ret_obj['data']['advanced']
        else:
            del ret_obj['data']['advanced']
    if signAlgorithm is not None:
        json_data['signAlgorithm'] = signAlgorithm
    elif 'signAlgorithm' in ret_obj['data']:
        if ret_obj['data']['signAlgorithm'] is not None:
            json_data['signAlgorithm'] = ret_obj['data']['signAlgorithm']
        else:
            del ret_obj['data']['signAlgorithm']
    if userIDToCreateSubject is not None:
        json_data['userIDToCreateSubject'] = userIDToCreateSubject
    elif 'userIDToCreateSubject' in ret_obj['data']:
        if ret_obj['data']['userIDToCreateSubject'] is not None:
            json_data['userIDToCreateSubject'] = ret_obj['data']['userIDToCreateSubject']
        else:
            del ret_obj['data']['userIDToCreateSubject']
    if scope is not None:
        json_data['scope'] = scope
    elif 'scope' in ret_obj['data']:
        if ret_obj['data']['scope'] is not None:
            json_data['scope'] = ret_obj['data']['scope']
        else:
            del ret_obj['data']['scope']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))
    if sorted_ret_obj != sorted_json_data:
        needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_put(
                "Update an existing openid entry", "{0}/{1}".format(uri, uuid),
                json_data, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def set(isvgAppliance, name, clientID, secret, certAlias, domains, issuerIdentifier, redirectToRPHostPort, userRealm, authUrl, tokenUrl, jwkUrl, logoffUrl, discoveryUrl, advanced, signAlgorithm="HS256", userIDToCreateSubject="sub", scope="openid", check_mode=False, force=False):
    """
    Creating or Modifying an openid entry
    """
    if (search(isvgAppliance, name))['data'] == {}:
        # Force the add - we already know object does not exist
        logger.info("Property {0} had no match, requesting to add new one.".format(name))
        return add(isvgAppliance, name, clientID, secret, certAlias, domains, issuerIdentifier, redirectToRPHostPort, userRealm, authUrl, tokenUrl, jwkUrl, logoffUrl, discoveryUrl, advanced, signAlgorithm, userIDToCreateSubject, scope, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Property {0} exists, requesting to update.".format(name))
        return update(isvgAppliance, name, clientID, secret, certAlias, domains, issuerIdentifier, redirectToRPHostPort, userRealm, authUrl, tokenUrl, jwkUrl, logoffUrl, discoveryUrl, advanced, signAlgorithm, userIDToCreateSubject, scope, check_mode=check_mode, force=force)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

requires_modules = ["mga", "federation"]
requires_version = None

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all SMTP server connections
    """
    return isamAppliance.invoke_get("Retrieving a list of all SMTP server connections",
                                    "/mga/server_connections/smtp/v1",
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieving a SMTP server connection
    """
    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieving a SMTP server connection",
                                        f"/mga/server_connections/smtp/{id}/v1",
                                        requires_modules=requires_modules, requires_version=requires_version)


def set(isamAppliance, name, connection, description='', locked=False, connectionManager=None, new_name=None, ignore_password_for_idempotency=False,
        check_mode=False, force=False):
    """
    Creating or Modifying an SMTP server connection
    """
    if _check_exists(isamAppliance, name=name) is False:
        # Force the add - we already know connection does not exist
        return add(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                   locked=locked, connectionManager=connectionManager, check_mode=check_mode, force=True)
    else:
        # Update request
        return update(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                      locked=locked, connectionManager=connectionManager, new_name=new_name, ignore_password_for_idempotency=ignore_password_for_idempotency,
                      check_mode=check_mode, force=force)


def add(isamAppliance, name, connection, description='', locked=False, connectionManager=None, check_mode=False,
        force=False):
    """
    Creating a SMTP server connection
    """
    if force is True or _check_exists(isamAppliance, name=name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a SMTP server connection",
                "/mga/server_connections/smtp/v1",
                _create_json(name=name, description=description, locked=locked, connection=connection,
                             connectionManager=connectionManager),
                             requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Deleting a SMTP server connection
    """
    if force is True or _check_exists(isamAppliance, name=name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = search(isamAppliance, name=name)
            id = ret_obj['data']
            return isamAppliance.invoke_delete(
                "Deleting a SMTP server connection",
                f"/mga/server_connections/smtp/{id}/v1",
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, connection, description='', locked=False, connectionManager=None, new_name=None, ignore_password_for_idempotency=False,
           check_mode=False, force=False):
    """
    Modifying a SMTP server connection

    Use new_name to rename the connection.
    """
    ret_obj = get(isamAppliance, name)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append(f"LDAP server connection {name} not found, skipping update.")
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        id = ret_obj["data"]["uuid"]

    needs_update = False

    json_data = _create_json(name=name, description=description, locked=locked, connection=connection, connectionManager=connectionManager)
    if new_name is not None:  # Rename condition
        json_data['name'] = new_name

    if force is not True:
        if 'uuid' in ret_obj['data']:
            del ret_obj['data']['uuid']
        if ignore_password_for_idempotency:
            if 'password' in connection:
                warnings.append("Request made to ignore password for idempotency check.")
                connection.pop('password', None)

        if not tools.json_equals(ret_obj, json_data):
            needs_update = True

    if force or needs_update:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Modifying a SMTP server connection",
                f"/mga/server_connections/smtp/{id}/v1", json_data,
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

def _create_json(name, description, locked, connection, connectionManager):
    """
    Create a JSON to be used for the REST API call
    """
    json = {
        "connection": connection,
        "type": "smtp",
        "name": name,
        "description": description,
        "locked": locked
    }
    # connection manager is optional
    if connectionManager is not None:
        json['connectionManager'] = connectionManager

    return json


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve UUID for named SMTP connection
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            ret_obj_new['data'] = obj['uuid']

    return ret_obj_new


def _check_exists(isamAppliance, name=None, id=None):
    """
    Check if SMTP Connection already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if (name is not None and obj['name'] == name) or (id is not None and obj['uuid'] == id):
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare SMTP Connections between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['uuid']
    for obj in ret_obj2['data']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import logging
import os.path

import ibmsecurity.utilities.tools
from ibmsecurity.utilities.tools import json_equals
from io import open

logger = logging.getLogger(__name__)

performCertCheck = True
try:
    from dateutil import parser
    #pip install python-dateutil
    import cryptography.hazmat.primitives.serialization.pkcs12
    import cryptography.x509
    from cryptography.x509 import Name, NameAttribute, NameOID
    from cryptography.x509.oid import ObjectIdentifier
    # pip install cryptography
except:
    performCertCheck = False


def get_all(isamAppliance, kdb_id, check_mode=False, force=False):
    """
    Retrieving personal certificate names and details in a certificate database
    """
    return isamAppliance.invoke_get(
        "Retrieving personal certificate names and details in a certificate database",
        f"/isam/ssl_certificates/{kdb_id}/personal_cert"
    )


def get(isamAppliance, kdb_id, cert_id, check_mode=False, force=False):
    """
    Retrieving a personal certificate from a certificate database
    """
    return isamAppliance.invoke_get(
        "Retrieving a personal certificate from a certificate database",
        f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}"
    )


def generate(isamAppliance, kdb_id, label, dn, expire='365', default='no', size='2048', signature_algorithm='',
             check_mode=False, force=False):
    """
    Generating a self-signed personal certificate in a certificate database
    """
    warnings = []

    if signature_algorithm is not None:
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "9.0.2.0") < 0:
            warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, signature_algorithm is not supported. Needs 9.0.2.0 or higher. Ignoring signature_algorithm for this call")
            json_obj = {
                "operation": 'generate',
                "label": label,
                "dn": dn,
                "expire": expire,
                'default': default,
                'size': size
            }
        else:
            json_obj = {
                "operation": 'generate',
                "label": label,
                "dn": dn,
                "expire": expire,
                'default': default,
                'size': size,
                'signature_algorithm': signature_algorithm
            }
    certexists, certsubject = _check(isamAppliance, kdb_id, label=label)
    if force or not certexists:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Generating a self-signed personal certificate in a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/personal_cert",
                json_obj,
                warnings=warnings
            )

    return isamAppliance.create_return_object(changed=False, warnings=warnings)


def rename(isamAppliance, kdb_id, cert_id, new_id,
           check_mode=False, force=False):
    """
    Rename a personal certificate.  New in 10.0.7
    """
    warnings = []

    if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.7.0") < 0:
        warnings.append(
            f"Appliance is at version: {isamAppliance.facts['version']}. Renaming a certificate requires at least 10.0.7.0"
        )
    else:
        certexists, certsubject = _check(isamAppliance, kdb_id, label=cert_id)
        newcertexists, nse = _check(isamAppliance, kdb_id, label=new_id)
        if force or (certexists and not newcertexists):
            if check_mode:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Renaming a personal certificate in a certificate database",
                    f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}",
                    {
                        'new_id': new_id
                    })


def set(isamAppliance, kdb_id, cert_id, default='no', check_mode=False, force=False):
    """
    Setting a personal certificate as default in a certificate database

    Obsolete since 10.0.3
    """
    warnings = []

    if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.3.0") > 0:
        warnings.append(
            f"Appliance is at version: {isamAppliance.facts['version']}. Setting certificates as default is no longer supported."
        )
    else:
        if force or _check_default(isamAppliance, kdb_id, cert_id, default):
            if check_mode:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Setting a personal certificate as default in a certificate database",
                    f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}",
                    {
                        'default': default
                    })

    return isamAppliance.create_return_object(warnings=warnings)


def receive(isamAppliance, kdb_id, label, cert, default='no', check_mode=False, force=False):
    """
    Receiving a personal certificate into a certificate database
    """
    certexists, certsubject = _check(isamAppliance, kdb_id, label=label)
    if force or not certexists:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Receiving a personal certificate into a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/personal_cert",
                [
                    {
                        'file_formfield': 'cert',
                        'filename': cert,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'default': default,
                    'operation': 'receive'
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, kdb_id, cert_id, check_mode=False, force=False, ignore_error=False):
    """
    Deleting a personal certificate from a certificate database
    """
    certexists, certsubject = _check(isamAppliance, kdb_id, label=cert_id)
    if force or certexists:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a personal certificate from a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}",
                ignore_error=ignore_error
            )

    return isamAppliance.create_return_object()


def export_cert(isamAppliance, kdb_id, cert_id, filename, check_mode=False, force=False):
    """
    Exporting a personal certificate from a certificate database
    """
    certexists, certsubject = _check(isamAppliance, kdb_id, label=cert_id)
    if force or (not certexists and not os.path.exists(filename)):
        if not check_mode:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Exporting a personal certificate from a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}?export",
                filename)

    return isamAppliance.create_return_object()


def import_cert(isamAppliance, kdb_id, cert, label=None, password=None, check_mode=False, force=False):
    """
    Importing a personal certificate into a certificate database
    Remark: you can add a label, but it's only used for a half-hearted check if the certificate already exists for versions < 11
            so it's NOT used to actually create the certificate !
    """
    warnings = []
    # certexists means that the SAME certificate exists
    certexists, certsubject = _check(isamAppliance, kdb_id, label=label, certificate=cert, password=password, simple_check=False)
    if certsubject is not None:
        logger.debug(f"Subject from new certificate {certsubject}")
    if force or not certexists:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            iviaVersion = isamAppliance.facts['version']
            post_data = {
                'file_formfield': 'cert',
                'filename': cert,
                'mimetype': 'application/octet-stream'
            }
            json_data = {
                'operation': 'import'
            }
            if password is not None:
                json_data['password'] = password
            if label is not None:
                if ibmsecurity.utilities.tools.version_compare(iviaVersion, "10.0.9.0") < 0:
                    warnings.append(f"Appliance at version: {iviaVersion}, label: {label} is not supported. Needs 10.0.9 or higher. Ignoring.")
                else:
                    logger.debug(f"Adding label: {label}")
                    json_data['label'] = label
            logger.debug(f"Posting json data: {json_data}")
            # This is a very normal error if you want to update an existing certificate
            # text: {"message":"CTGSK2021W A duplicate certificate already exists in the database."}
            #
            try:
                retObj = isamAppliance.invoke_post_files(
                             "Importing a personal certificate into a certificate database",
                             f"/isam/ssl_certificates/{kdb_id}/personal_cert",
                             [
                                 post_data
                             ],
                             json_data,
                             warnings=warnings
                         )
            except:
                # Delete the previous certificate and add the new.
                # The certificate exists with these values; so we want to overwrite it
                retObjDel = delete(isamAppliance, kdb_id, label, force=True, ignore_error=True)
                logger.debug(f"First try deleting {label}\n\n{retObjDel.get('rc', 200)}\n")
                if retObjDel.get("rc", 200) > 400:
                    # Delete using the subject as label
                    retObjDel = delete(isamAppliance, kdb_id, certsubject, force=True, ignore_error=True)
                    logger.debug(f"Second try at deleting using the new subject {certsubject}\n\n{retObjDel.get('rc', 200)}\n")
                # Try again
                retObj = isamAppliance.invoke_post_files(
                             "Retry Importing a personal certificate into a certificate database",
                             f"/isam/ssl_certificates/{kdb_id}/personal_cert",
                             [
                                 post_data
                             ],
                             json_data,
                             warnings=warnings
                         )
            return retObj

    return isamAppliance.create_return_object()


def extract_cert(isamAppliance, kdb_id, cert_id, password, filename, check_mode=False, force=False):
    """
    Extracting a personal certificate from a certificate database
    """
    if force or (_check(isamAppliance, kdb_id, label=cert_id) and not os.path.exists(filename)):
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = isamAppliance.invoke_post(
                "Extracting a personal certificate from a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}",
                {
                    'operation': 'extract',
                    'type': 'pkcs12',
                    'password': password
                })
            extracted_file = open(filename, 'wb')
            extracted_file.write(ret_obj['data'])
            extracted_file.close()

    return isamAppliance.create_return_object()


def _check(isamAppliance, kdb_id, label=None, certificate=None, password=None, simple_check=True):
    """
    Check if personal certificate already exists in certificate database
    This is idempotent now; using the subject of the incoming certificate (if you installed cryptography and dateutils
    Returns true or false and also the subject (in case of true)
    """
    if label is None:
        logger.debug("No label passed, so return false")
        if not performCertCheck:
            return False, None
        if simple_check:
            # This is to cater for current usage where the full idempotency is not requested
            return False, None

    if (not simple_check):
        if performCertCheck and (certificate is not None) and (password is not None):
            newCert = {}

            _enc = 'utf-8'
            with open(certificate, 'rb') as f:
                c = f.read()
            password = password.encode()
            p = cryptography.hazmat.primitives.serialization.pkcs12.load_pkcs12(c, password)

            x509 = p.cert.certificate

            # Define a mapping to fix some OIDs with "MAIL"
            attr_name_overrides = {
                ObjectIdentifier("0.9.2342.19200300.100.1.3"): "MAIL", # MAIL
                ObjectIdentifier("1.2.840.113549.1.9.1"): "MAIL",  # EMAIL_ADDRESS
                # Add other OIDs as needed
            }

            try:
                newCert['subject'] = x509.subject.rfc4514_string(attr_name_overrides=attr_name_overrides)
                newCert['issuer'] = x509.issuer.rfc4514_string(attr_name_overrides=attr_name_overrides)
            except:
                newCert['subject'] = x509.subject.rfc4514_string()
                newCert['issuer'] = x509.issuer.rfc4514_string()
                logger.info(
                    'Upgrade cryptography to (at least) version 36.0.0.  Install with pip install cryptography')

            newCert['notafter'] = x509.not_valid_after_utc.strftime("%Y-%m-%d")
            newCert['notbefore'] = x509.not_valid_before_utc.strftime("%Y-%m-%d")
            logger.debug(f"\nLOADING SUBJECT:\n {newCert['subject']}\n")
            ret_obj = get_all(isamAppliance, kdb_id)
            currentCert = None
            for certdb in ret_obj['data']:
                if certdb.get('id', '') == label:
                      currentCert = certdb
                elif certdb.get('id', '') == newCert['subject']:
                      currentCert = certdb
                elif certdb.get('subject', '') == newCert['subject']:
                      currentCert = certdb
            if currentCert is None:
                return False, newCert['subject']
            logger.debug(f"\nCurrent cert:\n {currentCert}\n")
            # remove values that are not checked
            currentCert.pop("keysize", None)
            currentCert.pop("version", None)

            # use date only to compare.  this simplifies the logic here, Python is not very strong with comparing dates in different timezones
            currentCert['notbefore'] = parser.parse(currentCert['notbefore'], ignoretz=True).strftime("%Y-%m-%d")
            currentCert['notafter'] = parser.parse(currentCert['notafter'], ignoretz=True).strftime("%Y-%m-%d")

            if json_equals(currentCert, newCert):
                # No updates needed
                return True, newCert['subject']
            else:
                return False, newCert['subject']
        else:
            logger.info(
                'Skipping management certificate full idempotency check, probably because cryptography>=36.0.0 is not available.  Install with pip install cryptography')

    # Do simple check
    ret_obj = get_all(isamAppliance, kdb_id)
    for certdb in ret_obj['data']:
        if certdb.get('id', '') == label:
           return True, certdb.get('subject', label)
    return False, None


def _check_default(isamAppliance, kdb_id, cert_id, default):
    """
    Check if personal certificate already exists in certificate database
    """
    ret_obj = get_all(isamAppliance, kdb_id)

    for certdb in ret_obj['data']:
        if certdb.get('id', '') == cert_id:
            if (certdb.get('default', 'false').lower() == 'true' and default.lower() == 'no') or (
                    certdb.get('default', 'false').lower() == 'false' and default.lower() == 'yes'):
                return True

    return False


def compare(isamAppliance1, isamAppliance2, kdb_id):
    """
    Compare signer certificates in certificate database between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, kdb_id)
    ret_obj2 = get_all(isamAppliance2, kdb_id)

    for cert in ret_obj1['data']:
        del cert['issuer']
        del cert['notafter']
        del cert['notafter_epoch']
        del cert['notbefore']
        del cert['notbefore_epoch']
        del cert['serial_number']
        del cert['sha1_fingerprint']
        del cert['subject']
    for cert in ret_obj2['data']:
        del cert['issuer']
        del cert['notafter']
        del cert['notafter_epoch']
        del cert['notbefore']
        del cert['notbefore_epoch']
        del cert['serial_number']
        del cert['sha1_fingerprint']
        del cert['subject']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2,
                                                    deleted_keys=['issuer', 'notafter', 'notafter_epoch', 'notbefore',
                                                                  'notbefore_epoch', 'serial_number',
                                                                  'sha1_fingerprint', 'subject'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve existing debug level.
    """
    return isamAppliance.invoke_get("Retrieving debug level for embedded ldap",
                                    "/isam/embedded_ldap/debug_level/v1")


def set(isamAppliance, trace=False, connection=False, search_filter=False, config_filter=False, acl_processing=False,
        statistics=False, statistics_entries=False, shell_backend=False, entry_parsing=False, sync_replication=False,
        uncategorized=False, check_mode=False, force=False):
    """
    Set debug levels in embedded ldap
    """
    if force is True or _check(isamAppliance, trace, connection, search_filter, config_filter, acl_processing,
                               statistics, statistics_entries, shell_backend, entry_parsing, sync_replication,
                               uncategorized) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Set debug level in embedded ldap",
                "/isam/embedded_ldap/debug_level/v1",
                {
                    "trace": trace,
                    "connection": connection,
                    "search.filter": search_filter,
                    "config.file": config_filter,
                    "acl.processing": acl_processing,
                    "statistics": statistics,
                    "statistics.entries": statistics_entries,
                    "shell.backend": shell_backend,
                    "entry.parsing": entry_parsing,
                    "sync.replication": sync_replication,
                    "uncategorized": uncategorized
                })

    return isamAppliance.create_return_object()


def _check(isamAppliance, trace=False, connection=False, search_filter=False, config_filter=False, acl_processing=False,
           statistics=False, statistics_entries=False, shell_backend=False, entry_parsing=False, sync_replication=False,
           uncategorized=False):
    new_trace = {
        "trace": trace,
        "connection": connection,
        "search.filter": search_filter,
        "config.file": config_filter,
        "acl.processing": acl_processing,
        "statistics": statistics,
        "statistics.entries": statistics_entries,
        "shell.backend": shell_backend,
        "entry.parsing": entry_parsing,
        "sync.replication": sync_replication,
        "uncategorized": uncategorized
    }

    ret_obj = get(isamAppliance)

    return ibmsecurity.utilities.tools.json_sort(new_trace) == ibmsecurity.utilities.tools.json_sort(ret_obj['data'])


def compare(isamAppliance1, isamAppliance2):
    """
    Compare debug levels in two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
# import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

# URI for this module
uri = "/wga/reverseproxy"


def get_all(isamAppliance, instance_id, check_mode=False, force=False):
    """
    Retrieving all statistics components and details - Reverse Proxy
    """
    try:
        return isamAppliance.invoke_get("Retrieving all statistics components - Reverse Proxy",
                                        "{0}/{1}/statistics".format(uri, instance_id),requires_model=requires_model)
    except:
        # Return empty array - exception thrown if list has no entries or does not exist
        ret_obj = isamAppliance.create_return_object()
        ret_obj['data'] = {}
        return ret_obj


def get_all_logs(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Retrieving all log files for a component - Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieving all statistics log files for a component - Reverse Proxy",
                                    "{0}/{1}/statistics/{2}/stats_files".format(uri,
                                                                                instance_id,
                                                                                component_id),requires_model=requires_model)


def get(isamAppliance, instance_id, component_id, file_id, options=None,
        size=None, start=None, check_mode=False, force=False):
    """
    Retrieving snippets of a statistics log file for a component - Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieving snippet of a statistics log file - Reverse Proxy",
                                    "{0}/{1}/statistics/{2}/stats_files/{3}".format(uri,
                                                                                    instance_id,
                                                                                    component_id,
                                                                                    file_id,
                                                                                    tools.create_query_string(
                                                                                        options=options, start=start,
                                                                                        size=size)),requires_model=requires_model)


def export_file(isamAppliance, instance_id, component_id, file_id, filename, check_mode=False, force=False):
    """
    Exporting a statistics log file for a component - Reverse Proxy
    """
    import os.path

    if force is True or (os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file("Exporting a Reverse Proxy statistics log file.",
                                                 "{0}/{1}/statistics/{2}/stats_files/{3}?export".format(uri,
                                                                                                        instance_id,
                                                                                                        component_id,
                                                                                                        file_id),
                                                 filename,requires_model=requires_model)
    return isamAppliance.create_return_object()


def set(isamAppliance, instance_id, component_id, status, hours, mins, secs,
        count, flush_interval, rollover_size, max_rollover_files, compress,
        check_mode=False, force=False):
    """
    Modify the statistics settings for a component
    """
    check_value, warnings = _check(isamAppliance,instance_id)

    if check_mode is True and check_value is True:
        return isamAppliance.create_return_object(changed=True,warnings=warnings)
    else:
        return isamAppliance.invoke_put(
            "Modify statistics settings for a component",
            "{0}/{1}/statistics/{2}".format(uri,
                                            instance_id,
                                            component_id),
            {
                'status': status,
                'interval_hours': hours,
                'interval_mins': mins,
                'interval_secs': secs,
                'count': count,
                'flush_interval': flush_interval,
                'rollover_size': rollover_size,
                'max_rollover_files': max_rollover_files,
                'compress': compress

            },requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)

def delete(isamAppliance, instance_id, component_id, file_id, check_mode=False, force=False):
    """
    Deleting the statistics log file or rollover file for a component - Reverse Proxy
    """
    check_value, warnings = _check(isamAppliance,instance_id)

    if force is False:
        try:
            ret_obj = get_default_snippet(isamAppliance, instance_id, component_id, file_id)
            delete_required = True  # Exception thrown if the file is empty
        except:
            delete_required = False

    if force is True or delete_required is True:
        if check_mode is True and check_value is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a statistics log file",
                "{0}/{1}/statistics/{2}/stats_files/{3}".format(uri,
                                                                instance_id,
                                                                component_id,
                                                                file_id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Deleting all the log files for a component - Reverse Proxy
    """
    check_value, warnings = _check(isamAppliance,instance_id)

    if force is False:
        try:
            ret_obj = get_all_logs(isamAppliance, instance_id, component_id)
            delete_required = True  # Exception thrown if the file is empty
        except:
            delete_required = False

    if force is True or delete_required is True:
        if check_mode is True and check_value is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Deleting all statistics log files",
                "{0}/{1}/statistics/{2}/stats_files".format(uri,
                                                            instance_id,
                                                            component_id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance,instance_id):
    """
    Check if it's appliance or not
    :param isamAppliance:
    :return: true|false, warnings message
    """
    ret_obj = get_all(isamAppliance,instance_id)
    check_value, warnings=False, ret_obj['warnings']

    if warnings == []:
        check_value = True
        return check_value, warnings
    else:
        return check_value, warnings



import logging
from ibmsecurity.isam.appliance import commit

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/capaths"
requires_modules = ['wga']
requires_version = None
import ibmsecurity.isam.web.kerberos_configuration.ca_paths as ca_paths


def add(isamAppliance, client_realm, server_realm, intermediate_realm, check_mode=False, force=False):
    """
    Add a client realm entry property
    """
    if ca_paths.search(isamAppliance, client_realm) == {}:
        return isamAppliance.create_return_object(
            warnings="Client Realm {0} not found, skipping add property: {1}.".format(client_realm, server_realm))

    ret_obj = ca_paths.get(isamAppliance, client_realm, check_mode=False, force=False)
    warnings = ret_obj["warnings"]

    # Check to see if server_realm is already present
    found_server_realm = False

    for obj in ret_obj['data']:
        if obj['name'] == server_realm:
            found_server_realm = True
            break

    if force is True or not found_server_realm:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if force is True and found_server_realm:
                return isamAppliance.invoke_put(
                    "Update a specified client realm property ",
                    "{0}/{1}/{2}".format(uri, client_realm, server_realm),
                    {"id": "capaths/{0}/{1}".format(client_realm, server_realm),
                     "value": intermediate_realm}, requires_modules=requires_modules,
                    requires_version=requires_version)
            else:
                return isamAppliance.invoke_post(
                    "Create an Kerberos Configuration Client realm entry property", "{0}/{1}".format(uri, client_realm),
                    {
                        "name": server_realm,
                        "value": intermediate_realm
                    }, requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, client_realm, server_realm, check_mode=False, force=False):
    """
    Delete an client realm entrie's particular property item
    """
    ret_obj = ca_paths.get(isamAppliance, client_realm, check_mode=check_mode, force=force)
    warnings = ret_obj["warnings"]

    if client_realm == {}:
        logger.info("Client Realm {0} not found, skipping delete.".format(client_realm))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # check if the server realm property to be deleted exists
            for obj in ret_obj['data']:
                if obj['name'] == server_realm:
                    return isamAppliance.invoke_delete(
                        "Delete an Kerberos Configuration Client Realm entries Property",
                        "{0}/{1}/{2}".format(uri, client_realm, server_realm), requires_modules=requires_modules,
                        requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, client_realm, server_realm, intermediate_realm, check_mode=False, force=False):
    """
    Update a specified Client Realm entry property, client_realm and server_realm passed in argument list needs to exist
    """
    if ca_paths.search(isamAppliance, client_realm) == {}:
        return isamAppliance.create_return_object(
            warnings="Client Realm {0} not found, skipping update.".format(client_realm))

    ret_obj = ca_paths.get(isamAppliance, client_realm)

    needs_update = False

    # check for existence of server_realm entry already

    found_server_realm = False

    for obj in ret_obj['data']:
        if obj['name'] == server_realm:
            found_server_realm = True
            break
    if found_server_realm:
        int_realm = ca_paths.get(isamAppliance, "{0}/{1}".format(client_realm, server_realm))['data']
        if intermediate_realm != int_realm[0]:
            needs_update = True
    else:
        return isamAppliance.create_return_object(
            warnings="Server Realm {0} to be updated under client realm:{1} not found, skipping update.".format(
                server_realm, client_realm))

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified client realm property ",
                "{0}/{1}/{2}".format(uri, client_realm, server_realm),
                {"id": "capaths/{0}/{1}".format(client_realm, server_realm),
                 "value": intermediate_realm}, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def set(isamAppliance, client_realm, server_realm=None, intermediate_realm=None, check_mode=False, force=False):
    """
        Creating or Modifying an Kerberos Configuration client realm Property,
    """
    # check for an input of intermediate realm if server realm being set is not None

    if not server_realm is None:
        if intermediate_realm is None:
            return isamAppliance.create_return_object(
                warnings="a non null Intermediate Realm needs to be passed as input, when Server realm :{0} is being set".format(
                    server_realm))

    # check if client realm exists, if it does not exit with warning message
    if ca_paths.search(isamAppliance, name=client_realm)['data'] == {}:
        return isamAppliance.create_return_object(
            warnings="Client_realm: {0}, does not exist".format(client_realm))
    else:
        logger.info("Client Realm {0} exists, requesting to update its property.".format(client_realm))
        ret_obj = ca_paths.get(isamAppliance, client_realm)
        found_server_realm = False

        for obj in ret_obj['data']:
            if obj['name'] == server_realm:
                found_server_realm = True
                break
        if found_server_realm is False:
            return add(isamAppliance, client_realm, server_realm, intermediate_realm, check_mode=check_mode,
                       force=force)
        else:
            return update(isamAppliance, client_realm, server_realm, intermediate_realm, check_mode=check_mode,
                          force=force)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/sts/templates"
requires_modules = ['federation']
requires_version = "9.0.1.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of STS chain templates
    """
    return isamAppliance.invoke_get("Retrieve a list of STS chain templates", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search STS Chain by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found STS Chain Template {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific STS chain template
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings = ["STS Chain Template {0} had no match, skipping retrieval.".format(name)]
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    """
    Internal function to get data using "id" - used to avoid extra calls
    """
    return isamAppliance.invoke_get("Retrieve a specific STS chain template", "{0}/{1}".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def set(isamAppliance, name, chainItems=[], description=None, new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying an STS Chain
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know Chain does not exist
        logger.info("STS Chain template {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name=name, chainItems=chainItems, description=description, check_mode=check_mode,
                   force=True)
    else:
        # Update request
        logger.info("STS Chain template {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name=name, chainItems=chainItems, description=description, new_name=new_name,
                      check_mode=check_mode, force=force)


def _flatten_chain(chainItems):
    """
    Convert chainItems from JSON to a simple string for use in comparison, ignore prefix if any.

    DO NOT Sort the JSON containing chainItems for comparison. Sequence is important! Use this function instead.
    """
    chainItemsString = ''
    for ci in chainItems:
        chainItemsString += ci['id'] + ":" + ci['mode'] + " "

    logger.debug("Flattened String: {0}".format(chainItemsString))

    return chainItemsString


def add(isamAppliance, name, chainItems=[], description=None, check_mode=False, force=False):
    """
    Create an STS chain template
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "chainItems": chainItems
            }
            if description is not None:
                json_data['description'] = description
            return isamAppliance.invoke_post(
                "Create an STS chain template", uri, json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a specific STS chain template
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    chain_id = ret_obj['data']

    if chain_id == {}:
        logger.info("STS Chain Template {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a specific STS chain template",
                "{0}/{1}".format(uri, chain_id),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, chainItems=[], description=None, new_name=None, check_mode=False, force=False):
    """
    Update a specific STS chain
    """
    warnings = []
    chain_id, update_required, json_data = _check(isamAppliance, name, chainItems, description, new_name)
    if chain_id is None:
        warnings.append("Cannot update data for unknown STS Chain Template: {0}".format(name))
    else:
        if force is True or update_required is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Update a specific STS chain template",
                    "{0}/{1}".format(uri, chain_id), json_data,
                    requires_modules=requires_modules,
                    requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, name, chainItems, description, new_name=None):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {
        "name": name,
        "chainItems": chainItems
    }
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("STS Chain template not found, returning no update required.")
        return None, update_required, json_data
    else:
        chain_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        if description is not None:
            json_data['description'] = description
        if (new_name is not None and ret_obj['data']['name'] != name) or (
                description is not None and ret_obj['data']['description'] != description) or (
                _flatten_chain(chainItems) != _flatten_chain(ret_obj['data']['chainItems'])):
            logger.info("Changes detected, update needed.")
            update_required = True
        else:
            logger.debug("No changes detected.")

    return chain_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare STS Chains Templates between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        for ci in obj['chainItems']:
            del ci['prefix']
    for obj in ret_obj2['data']:
        del obj['id']
        for ci in obj['chainItems']:
            del ci['prefix']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'chainItems/prefix'])



import logging
import json
import ibmsecurity.isam.aac.server_connections.ws
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the complete list of SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the complete list of SCIM configuration settings",
                                    "/mga/scim/configuration", requires_modules=requires_modules,
                                    requires_version=requires_version)


def get_user_profile(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current user profile SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current user profile SCIM configuration settings",
                                    "/mga/scim/configuration/urn:ietf:params:scim:schemas:core:2.0:User",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get_ent_profile(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current enterprise user profile SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current enterprise user profile SCIM configuration settings",
                                    f"{uri}/urn:ietf:params:scim:schemas:extension:enterprise:2.0:User",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get_isam_user(isamAppliance, check_mode=False, force=False):
    """
    Retrieve configuration of SCIM ISAM user settings
    """
    return isamAppliance.invoke_get("Retrieve configuration of SCIM ISAM user settings",
                                    "/mga/scim/configuration/urn:ietf:params:scim:schemas:extension:isam:1.0:User",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)

def get_ext_auth_service_config(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current external authentication service SCIM configuration
    """
    return isamAppliance.invoke_get("Retrieving the current external authentication service SCIM configuration ",
                                    f"{uri}/urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def get_general_config(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current general SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current general SCIM configuration settings",
                                    f"{uri}/general",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def get_group_config(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current group SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current group SCIM configuration settings",
                                    f"{uri}/urn:ietf:params:scim:schemas:core:2.0:Group",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def get_ldap_objs(isamAppliance, ldap_connection, check_mode=False, force=False):
    """
    Retrieving the current list of ldap object classes from the configured ldap
    """
    return isamAppliance.invoke_get("Retrieving the current list of ldap object classes from the configured ldap",
                                    "{0}/urn:ietf:params:scim:schemas:core:2.0:User/ldap_objectclasses?ldap_connection={1}".format(
                                        uri, ldap_connection),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def get_ldap_attrs(isamAppliance, ldap_connection, ldap_objectclasses='', check_mode=False, force=False):
    """
    Retrieving the current list of user associated ldap attributes from the configured
    """
    return isamAppliance.invoke_get(
        "Retrieving the current list of user associated ldap attributes from the configured ",
        "{0}/urn:ietf:params:scim:schemas:core:2.0:User/ldap_attributes?ldap_connection={1}&ldap_objectclasses={1}".format(
            uri, ldap_connection, ldap_objectclasses),
        requires_modules=requires_modules,
        requires_version=requires_version
        )

def update_user_profile(isamAppliance, ldap_connection, user_suffix, search_suffix, check_mode=False, force=False):
    """
    Updating the user profile SCIM configuration settings
    """
    ret_obj = get_user_profile(isamAppliance)
    ret_obj = ret_obj['data']['urn:ietf:params:scim:schemas:core:2.0:User']
    del ret_obj['ldap_connection']
    del ret_obj['user_suffix']
    del ret_obj['search_suffix']

    ret_obj['ldap_connection'] = ldap_connection
    ret_obj['user_suffix'] = user_suffix
    ret_obj['search_suffix'] = search_suffix
    return isamAppliance.invoke_put(
        "Updating the user profile SCIM configuration settings",
        "/mga/scim/configuration/urn:ietf:params:scim:schemas:core:2.0:User",
        ret_obj, requires_modules=requires_modules,
        requires_version=requires_version)


def update_isam_user(isamAppliance, ldap_connection=None, isam_domain=None, update_native_users=None, connection_type=None, attrs_dir=None, enforce_password_policy=None, check_mode=False, force=False):
    """
    Updating the ISAM user SCIM configuration settings
    """
    ret_obj = get_isam_user(isamAppliance)
    ret_obj = ret_obj['data']['urn:ietf:params:scim:schemas:extension:isam:1.0:User']

    new_obj = {}

    update_required = False

    if ldap_connection is None:
        if 'ldap_connection' in ret_obj:
            new_obj['ldap_connection'] = ret_obj['ldap_connection']
    else:
        new_obj['ldap_connection'] = ldap_connection
        if 'ldap_connection' in ret_obj:
            if ret_obj['ldap_connection'] != ldap_connection:
                update_required = True
        else:
            update_required = True

    if isam_domain is None:
        if 'isam_domain' in ret_obj:
            new_obj['isam_domain'] = ret_obj['isam_domain']
    else:
        new_obj['isam_domain'] = isam_domain
        if 'isam_domain' in ret_obj:
            if ret_obj['isam_domain'] != isam_domain:
                update_required = True
        else:
            update_required = True

    if update_native_users is None:
        if 'update_native_users' in ret_obj:
            new_obj['update_native_users'] = ret_obj['update_native_users']
    else:
        new_obj['update_native_users'] = update_native_users
        if 'update_native_users' in ret_obj:
            if ret_obj['update_native_users'] != update_native_users:
                update_required = True
        else:
            update_required = True

    if connection_type is None:
        if 'connection_type' in ret_obj:
            new_obj['connection_type'] = ret_obj['connection_type']
    else:
        new_obj['connection_type'] = connection_type
        if 'connection_type' in ret_obj:
            if ret_obj['connection_type'] != connection_type:
                update_required = True
        else:
            update_required = True

    if attrs_dir is None:
        if 'attrs_dir' in ret_obj:
            new_obj['attrs_dir'] = ret_obj['attrs_dir']
    else:
        new_obj['attrs_dir'] = attrs_dir
        if 'attrs_dir' in ret_obj:
            if ret_obj['attrs_dir'] != attrs_dir:
                update_required = True
        else:
            update_required = True

    if enforce_password_policy is None:
        if 'enforce_password_policy' in ret_obj:
            new_obj['enforce_password_policy'] = ret_obj['enforce_password_policy']
    else:
        new_obj['enforce_password_policy'] = enforce_password_policy
        if 'enforce_password_policy' in ret_obj:
            if ret_obj['enforce_password_policy'] != enforce_password_policy:
                update_required = True
        else:
            update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating the ISAM user SCIM configuration settings",
                "/mga/scim/configuration/urn:ietf:params:scim:schemas:extension:isam:1.0:User",
                new_obj, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(changed=False)




def update_ent_profile(isamAppliance, mappings, check_mode=False, force=False):
    """
    Updating the enterprise user profile SCIM configuration settings

    """
    return isamAppliance.invoke_put("Updating the enterprise user profile SCIM configuration settings",
                                    f"{uri}/urn:ietf:params:scim:schemas:extension:enterprise:2.0:User",
                                    mappings,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def update_ext_auth_service_config(isamAppliance, schemas, check_mode=False, force=False):
    """
    Updating the external authentication service SCIM configuration settings
    """
    current_objs = get_ext_auth_service_config(isamAppliance)
    current_configs = current_objs['data']['urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS']
    ws_connections = ibmsecurity.isam.aac.server_connections.ws.get_all(isamAppliance)
    ws_connections = ws_connections['data']
    update_required = False

    for obj in schemas:
        found = False
        for config in current_configs:
            if obj['connection'] == config['connection']:
                found = True
                new_schemas = obj['schemas']
                old_schemas = config['schemas']
                sorted_new = json_sort(new_schemas)
                sorted_old = json_sort(old_schemas)
                if sorted_new != sorted_old:
                    update_required = True

        if found is False:
            exist = False
            for ws in ws_connections:
                if obj['connection'] == ws['uuid']:
                    update_required = True
                    exist = True
            if exist is False:
                warnings = f"Did not find connection {obj['connection']} in the configured server list."
                return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Updating the external authentication service SCIM configuration settings",
                                            f"{uri}/urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS",
                                            schemas,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version
                                            )

    return isamAppliance.create_return_object(changed=False)


def set_all(isamAppliance, settings, check_mode=False, force=False):
    """
    Update entire SCIM settings
    """
    if settings is None or settings == '':
        return isamAppliance.create_return_object(
            warnings=["Need to pass content for scim configuration"])
    else:
        # Feature: Converting python string to dict (if required)
        # Attention: JSON strings must use " quotes according to RFC 8259
        # Example: '{"a":1, "b": 2, "c": 3}'
        if isinstance(settings, str):
            settings = json.loads(settings)
        if force or not _check(isamAppliance, settings):
            if check_mode:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Update SCIM settings",
                    "/mga/scim/configuration",
                    settings)

    return isamAppliance.create_return_object()

def _check(isamAppliance, settings):
    """
    Check if scim configuration is identical with server
    """
    ret_obj = get_all(isamAppliance)
    logger.debug("Comparing server scim configuration with desired configuration.")
    # Converting python ret_obj['data'] and settings from type dict to valid JSON (RFC 8259)
    # e.g. converts python boolean 'True' -> to JSON literal lowercase value 'true'
    cur_json_string = json.dumps(ret_obj['data'])
    cur_sorted_json = json_sort(cur_json_string)
    logger.debug(f"Server JSON : {cur_sorted_json}")
    given_json_string = json.dumps(settings)
    given_sorted_json = json_sort(given_json_string)
    logger.debug(f"Desired JSON: {given_sorted_json}")
    if cur_sorted_json != given_sorted_json:
        return False
        logger.debug("Changes detected!")
    else:
        logger.debug("Server configuration is identical with desired configuration. No change necessary.")
        return True


def update_group_config(isamAppliance, ldap_object_classes, group_dn, check_mode=False, force=False):
    """
    Updating the group SCIM configuration settings
    """
    obj = get_group_config(isamAppliance)
    obj = obj['data']['urn:ietf:params:scim:schemas:core:2.0:Group']

    current_dn = obj['group_dn']
    current_classes = obj['ldap_object_classes']

    if current_dn == group_dn:
        sorted_new = json_sort(ldap_object_classes)
        sorted_old = json_sort(current_classes)
        if sorted_new == sorted_old:
            update_required = False
        else:
            update_required = True
    else:
        update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            data = {}
            data['group_dn'] = group_dn
            data['ldap_object_classes'] = ldap_object_classes
            return isamAppliance.invoke_put("Updating the group SCIM configuration settings",
                                            f"{uri}/urn:ietf:params:scim:schemas:core:2.0:Group",
                                            data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version
                                            )

    return isamAppliance.create_return_object(changed=False)


def update_mode(isamAppliance, schema_name, scim_attribute, mode, scim_subattribute=None, check_mode=False,
                force=False):
    """
    Updating the mode of a SCIM attribute
    """

    ret_obj = get_general_config(isamAppliance)

    mode = mode.lower()

    objs = ret_obj['data']['attribute_modes']

    update_required = True

    if scim_subattribute is None:
        obj1 = {'mode': mode, 'attribute': scim_attribute}
    else:
        obj1 = {'mode': mode, 'attribute': scim_attribute, 'subattribute': scim_subattribute}

    obj1 = json_sort(obj1)

    for obj in objs:
        schema = obj['schema']
        if schema == schema_name:
            modes = obj['modes']
            for anitem in modes:
                obj2 = json_sort(anitem)
                if obj1 == obj2:
                    update_required = False

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            if scim_subattribute is None:
                return isamAppliance.invoke_put("Updating the mode of a SCIM attribute",
                                                "{0}/general/attribute_modes/{1}/{2}".format(uri, schema_name,
                                                                                             scim_attribute),
                                                {'mode': mode},
                                                requires_modules=requires_modules,
                                                requires_version=requires_version
                                                )
            else:
                return isamAppliance.invoke_put("Updating the mode of a SCIM attribute",
                                                "{0}/general/attribute_modes/{1}/{2}/{3}".format(uri, schema_name,
                                                                                                 scim_attribute,
                                                                                                 scim_subattribute),
                                                {'mode': mode},
                                                requires_modules=requires_modules,
                                                requires_version=requires_version
                                                )

    return isamAppliance.create_return_object(changed=False)


def reset_mode(isamAppliance, schema_name, scim_attribute, scim_subattribute=None, check_mode=False, force=False):
    """
    Resetting a SCIM attribute mode to default
    """
    ret_obj = get_general_config(isamAppliance)

    objs = ret_obj['data']['attribute_modes']

    update_required = False

    for obj in objs:
        schema = obj['schema']
        if schema == schema_name:
            modes = obj['modes']
            for anitem in modes:
                if anitem['attribute'] == scim_attribute:
                    if scim_subattribute is not None:
                        if 'subattribute' in anitem:
                            if anitem['subattribute'] == scim_subattribute:
                                update_required = True
                    elif ('subattribute' in anitem) is False:
                        update_required = True


    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if scim_subattribute is None:
                return isamAppliance.invoke_delete("Resetting a SCIM attribute mode to default",
                                               "{0}/general/attribute_modes/{1}/{2}".format(uri, schema_name,
                                                                                            scim_attribute),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version
                                               )
            else:
                return isamAppliance.invoke_delete("Resetting a SCIM attribute mode to default",
                                               "{0}/general/attribute_modes/{1}/{2}/{3}".format(uri, schema_name,
                                                                                                scim_attribute,
                                                                                                scim_subattribute),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version
                                               )

    return isamAppliance.create_return_object(changed=False)



import logging
import ibmsecurity.utilities.tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/authzserver"
requires_model = "Appliance"
docker_warning = ['API invoked requires model: Appliance, appliance is of deployment model: Docker.']


def get_all(isamAppliance, id, stanza_id, check_mode=False, force=False):
    """
    Retrieving all configuration entries for a stanza - Authorization Server
    """
    try:
        return isamAppliance.invoke_get(
            description="Retrieving all configuration entries for a stanza - Authorization Server",
            uri="{0}/{1}/configuration/stanza/{2}/v1".format(uri,
                                                             id,
                                                             stanza_id), requires_model=requires_model)
    except:
        # Return empty array - exception thrown if stanza has no entries or does not exist
        ret_obj = isamAppliance.create_return_object()
        ret_obj['data'] = {}
        return ret_obj


def get(isamAppliance, id, stanza_id, entry_id, check_mode=False, force=False):
    """
    Retrieving a specific configuration entry - Authorization Server
    """
    return isamAppliance.invoke_get(description="Retrieving a specific configuration entry - Authorization Server",
                                    uri="{0}/{1}/configuration/stanza/{2}/entry_name/{3}/v1".format(uri,
                                                                                                    id, stanza_id,
                                                                                                    entry_id),
                                    requires_model=requires_model)


def add(isamAppliance, id, stanza_id, entries, check_mode=False, force=False):
    """
    Add configuration entry by stanza - Authorization Server
    """
    if _isDocker(isamAppliance, id, stanza_id):
        return isamAppliance.create_return_object(warnings=docker_warning)

    if isinstance(entries, basestring):
        import ast
        entries = ast.literal_eval(entries)

    add_required = False

    if force is False:
        add_entries = []
        for entry in entries:
            exists, update_required, value = _check(isamAppliance, id, stanza_id, entry[0], entry[1])
            if exists is True:
                logger.debug(
                    'Entries exists {0}/{1}/{2}/{3}! Will be ignored.'.format(id, stanza_id, entry[0],
                                                                              entry[1]))
            else:
                add_entries.append(entry)
                add_required = True
        entries = add_entries

    if force is True or add_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, id, stanza_id, entries)

    return isamAppliance.create_return_object()


def _add(isamAppliance, id, stanza_id, entries):
    return isamAppliance.invoke_post(
        description="Add configuration entry by stanza - Authorization Server",
        uri="{0}/{1}/configuration/stanza/{2}/entry_name/v1".format(uri, id, stanza_id),
        data={"entries": entries}, requires_model=requires_model)


def _isDocker(isamAppliance, id, stanza_id):
    ret_obj = get_all(isamAppliance, id, stanza_id)
    warnings = ret_obj['warnings']
    if warnings and 'Docker' in warnings[0]:
        return True
    else:
        return False


def set(isamAppliance, id, stanza_id, entries, check_mode=False, force=False):
    """
    Set a configuration entry or entries by stanza - Authorization Server

    Note: entries has to be [['key', 'value1'], ['key', 'value2]], cannot provide [['key', ['value1', 'value2']]]
    get() returns the second format - thus lots of logic to handle this discrepancy.

    Smart enough to update only that which is needed.
    """
    if _isDocker(isamAppliance, id, stanza_id):
        return isamAppliance.create_return_object(warnings=docker_warning)

    if isinstance(entries, basestring):
        import ast
        entries = ast.literal_eval(entries)

    set_update = False
    set_entries = []

    if force is False:
        for entry in _collapse_entries(entries):
            process_entry = False
            exists, update_required, cur_value = _check(isamAppliance, id, stanza_id, entry[0], entry[1])
            if exists is False:
                set_update = True
                process_entry = True
            elif update_required is True:
                set_update = True
                process_entry = True
                for val in cur_value:
                    # Force delete of existing values, new values will be added
                    delete(isamAppliance, id, stanza_id, entry[0], val, check_mode, True)
                    logger.info(
                        'Deleting entry, will be re-added: {0}/{1}/{2}/{3}'.format(id, stanza_id, entry[0],
                                                                                   val))
            if process_entry is True:
                if isinstance(entry[1], list):
                    for v in entry[1]:
                        set_entries.append([entry[0], v])
                else:
                    set_entries.append([entry[0], entry[1]])

    if force is True or set_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, id, stanza_id, set_entries)

    return isamAppliance.create_return_object()


def _collapse_entries(entries):
    """
    Convert [['key', 'value1'], ['key', 'value2]] to [['key', ['value1', 'value2']]]
    Expect key values to be consecutive. Will maintain order as provided.
    """
    if entries is None or len(entries) < 1:
        return []
    else:
        cur_key = entries[0][0]
        cur_value = []
        new_entry = []

    for entry in entries:
        if entry[0] == cur_key:
            cur_value.append(entry[1])
        else:
            new_entry.append([cur_key, cur_value])
            # reset current key
            cur_key = entry[0]
            cur_value = [entry[1]]

    new_entry.append([cur_key, cur_value])

    return new_entry


def delete(isamAppliance, id, stanza_id, entry_id, value_id='', check_mode=False, force=False):
    """
    Deleting a value from a configuration entry - Authorization Server
    """
    if _isDocker(isamAppliance, id, stanza_id):
        return isamAppliance.create_return_object(warnings=docker_warning)

    exists = False
    if force is False:
        exists, update_required, value = _check(isamAppliance, id, stanza_id, entry_id, value_id)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            # URL being encoded primarily to handle request-log-format that has "%" values in them
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}/value/{4}/v1".format(uri, id,
                                                                                          stanza_id, entry_id, value_id)
            # Replace % with %25 if it is not encoded already
            import re
            ruri = re.sub("%(?![0-9a-fA-F]{2})", "%25", f_uri)
            # URL encode
            import urllib.parse
            full_uri = urllib.parse.quote(ruri)
            return isamAppliance.invoke_delete(
                description="Deleting a value from a configuration entry - Authorization Server",
                uri=full_uri, requires_model=requires_model)

    return isamAppliance.create_return_object()


def delete_all(isamAppliance, id, stanza_id, entry_id, check_mode=False, force=False):
    """
    Deleting all values from a configuration entry - Authorization Server
    """

    delete_required = False
    if force is False:
        try:
            ret_obj = get(isamAppliance, id, stanza_id, entry_id)
            warnings = ret_obj['warnings']
            if warnings and 'Docker' in warnings[0]:
                return isamAppliance.create_return_object(warnings=ret_obj['warnings'])

            if ret_obj['data'] != {}:
                delete_required = True
        except:
            pass

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}/v1".format(uri, id, stanza_id, entry_id)
            return isamAppliance.invoke_delete(
                description="Deleting all values from a configuration entry - Authorization Server", uri=f_uri,
                requires_model=requires_model)

    return isamAppliance.create_return_object()


def update(isamAppliance, id, stanza_id, entry_name_id, value_id, check_mode=False, force=False):
    """
    Updating a configuration entry or entries by stanza - Authorization Server
    """
    if _isDocker(isamAppliance, id, stanza_id):
        return isamAppliance.create_return_object(warnings=docker_warning)

    if force is False:
        exists, update_required, cur_value = _check(isamAppliance, id, stanza_id, entry_name_id, value_id)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                description="Updating a configuration entry or entries by stanza - Authorization Server",
                uri="{0}/{1}/configuration/stanza/{2}/entry_name/{3}/v1".format(uri,
                                                                                id,
                                                                                stanza_id,
                                                                                entry_name_id),
                data={
                    'value': value_id
                }, requires_model=requires_model)

    return isamAppliance.create_return_object()


def _check(isamAppliance, id, stanza_id, entry_id, value_id):
    """
    Check if entry/value exists and if exists then if update is required
    """
    try:
        ret_obj = get(isamAppliance, id, stanza_id, entry_id)
        exists = True
        update_required = False
        value = ret_obj['data'][entry_id]
    except:
        return False, True, None  # Exception means entry / stanza not found

    logger.info("Entry found in acld:{0}, stanza:{1}, entryid:{2}, value:{3}".format(id,
                                                                                     stanza_id,
                                                                                     entry_id,
                                                                                     value))
    logger.debug("Existing Value(s): {0}".format(value))
    logger.debug("Value to update  : {0}".format(value_id))

    if isinstance(value_id, list):
        if value != value_id:  # Comparing list with no sorting... sequence of values is of importance
            logger.debug("Value arrays do not match!")
            update_required = True
    else:  # assuming base string provided for value_id
        if len(value) == 1:
            if str(value_id) != str(value[0]):
                logger.debug("Single value do not match!")
                update_required = True
                exists = False  # to satisfy delete call
        else:  # base string will not match a zero length array or multiple values in it
            logger.debug("Current non-single value does not match provided single value!")
            update_required = True
            if value_id not in value:
                logger.debug("Current non-single value does not contain provided single value!")
                exists = False

    return exists, update_required, value


def compare(isamAppliance1, isamAppliance2, id, stanza_id, id2=None):
    """
    Compare stanza/entries in two appliances authorization server configuration
    """
    if id2 is None or id2 == '':
        id2 = id

    ret_obj1 = get_all(isamAppliance1, id, stanza_id)
    ret_obj2 = get_all(isamAppliance2, id2, stanza_id)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/alias_service"
requires_modules = ["federation"]
requires_version = "9.0.1.0"


def get(isamAppliance, sortBy=None, count=None, start=None, filter=None, check_mode=False, force=False):
    """
    Retrieve alias associations
    """
    return isamAppliance.invoke_get("Retrieve alias associations", "{0}{1}".format(uri, tools.create_query_string(
        sortBy=sortBy, count=count, start=start, filter=filter)), requires_modules=requires_modules,
                                    requires_version=requires_version)


def add(isamAppliance, username, federation_id, aliases, type=None, partner_id=None, check_mode=False, force=False):
    """
    Create an alias association

    TODO: Need to understand uniqueness of federation/partner_id and username to complete this
    """
    warnings = ["Idempotency has not been coded for this function."]
    if force is True or _check(isamAppliance, username) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            fed_id = federation_id
            if partner_id is not None:
                fed_id = fed_id + "|" + partner_id
            json_data = {
                "username": username,
                "federation_id": fed_id,
                "aliases": aliases
            }
            if type is not None:
                json_data['type'] = type
            return isamAppliance.invoke_post(
                "Create an alias association", uri, json_data, warnings=warnings,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, username):
    """
    TODO: Code the idempotency check logic
    """
    return False



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve available updates
    """
    return isamAppliance.invoke_get("Retrieving available updates",
                                    "/updates/available.json")


def discover(isamAppliance, check_mode=False, force=False):
    """
    Discover available updates
    """
    return isamAppliance.invoke_get("Discover available updates",
                                    "/updates/available/discover")


def upload(isamAppliance, file, check_mode=False, force=False):
    """
    Upload Available Update
    """
    if force is True or _check_file(isamAppliance, file) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Upload Available Update",
                "/core/updates/available",
                [{
                    'file_formfield': 'uploadedfile',
                    'filename': file,
                    'mimetype': 'application/octet-stream'
                }],
                {}, json_response=False)

    return isamAppliance.create_return_object()


def _check_file(isamAppliance, file):
    """
    Parse the file name to see if it is already uploaded - use version and release date from pkg file name
    Also check to see if the firmware level is already uploaded
    Note: Lot depends on the name of the file.

    :param isamAppliance:
    :param file:
    :return:
    """
    import os.path

    # If there is an exception then simply return False
    # Sample filename - isam_9.0.2.0_20161102-2353.pkg
    logger.debug(f"Checking provided file is ready to upload: {file}")
    try:
        # Extract file name from path
        f = os.path.basename(file)
        fn = os.path.splitext(f)
        logger.debug(f"File name without path: {fn[0]}")

        # Split of file by '-' hyphen and '_' under score
        import re
        fp = re.split('-|_', fn[0])
        firm_file_version = fp[1]
        firm_file_product = fp[0]
        firm_file_date = fp[2]
        logger.debug(f"PKG file details: {firm_file_product}: version: {firm_file_version} date: {firm_file_date}")

        # Check if firmware level already contains the update to be uploaded or greater, check Active partition
        # firmware "name" of format - isam_9.0.2.0_20161102-2353
        import ibmsecurity.isam.base.firmware
        ret_obj = ibmsecurity.isam.base.firmware.get(isamAppliance)
        for firm in ret_obj['data']:
            # Split of file by '-' hyphen and '_' under score
            fp = re.split('-|_', firm['name'])
            firm_appl_version = fp[1]
            firm_appl_product = fp[0]
            firm_appl_date = fp[2]
            logger.debug(f"Partition details ({firm['partition']}): {firm_appl_product}: version: {firm_appl_version} date: {firm_appl_date}")
            if firm['active'] is True:
                from ibmsecurity.utilities import tools
                if tools.version_compare(firm_appl_version, firm_file_version) >= 0:
                    logger.info(
                        f"Active partition has version {firm_appl_version} which is greater or equals than install package at version {firm_file_version}.")
                    return True
                else:
                    logger.info(
                        f"Active partition has version {firm_appl_version} which is smaller than install package at version {firm_file_version}.")

        # Check if update uploaded - will not show up if installed though
        ret_obj = get(isamAppliance)
        for upd in ret_obj['data']:
            rd = upd['release_date']
            rd = rd.replace('-', '')  # turn release date into 20161102 format from 2016-11-02
            if upd['version'] == fp[1] and rd == fp[2]:  # Version of format 9.0.2.0
                return True
    except Exception as e:
        logger.debug(f"Exception occured: {e}")
        pass

    return False


def install(isamAppliance, type, version, release_date, name, check_mode=False, force=False):
    """
    Install Available Update
    """
    if force is True or _check(isamAppliance, type, version, release_date, name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = isamAppliance.invoke_post("Install Available Update",
                                                "/updates/available/install",
                                                {"updates": [
                                                    {
                                                        "type": type,
                                                        "version": version,
                                                        "release_date": release_date,
                                                        "name": name
                                                    }
                                                ]
                                                })
            isamAppliance.facts['version'] = version
            return ret_obj

    return isamAppliance.create_return_object()

def _check(isamAppliance, type, version, release_date, name):
    ret_obj = get(isamAppliance)
    warnings = []
    for upd in ret_obj['data']:
        # If there is an installation in progress then abort
        # API changed in v10.0.5.0 , state, schedule_date,
        #   iso_scheduled_date and expired_install are no longer available.
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.5.0") >= 0:
            if 'state' in upd and upd['state'] == 'Installing':
                logger.debug("Detecting a state of installing...")
                return False
        else:
            warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, state can no longer be checked in 10.0.5.0 or higher. Ignoring for this call.")

        logger.info(upd)

        if upd['type'] == type and upd['version'] == version and \
                upd['release_date'] == release_date and upd['name'] == name:
            logger.debug("Requested firmware ready for install...")
            return True

    logger.debug("Requested firmware not available for install...")

    return False



import logging
import ibmsecurity.utilities.tools
import json

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get management authorization - roles
    """
    return isamAppliance.invoke_get("Get management authorization - roles",
                                    "/authorization/roles/v1")


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Get management authorization role
    """
    return isamAppliance.invoke_get("Get management authorization role",
                                    f"/authorization/roles/{name}/v1")


def _check(isamAppliance, name):
    """
    Check if management authorization role exists
    """
    ret_obj = get_all(isamAppliance)

    for role in ret_obj['data']:
        if role['name'] == name:
            return True

    return False

def add(isamAppliance, name, users=None, groups=None, features=None, check_mode=False, force=False):
    """
    Add a management authorization role
    """
    if force is True or _check(isamAppliance, name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Add management authorization role",
                                             "/authorization/roles/v1",
                                             {
                                                 'name': name,
                                                 'users': users,
                                                 'groups': groups,
                                                 'features': features
                                             })

    return isamAppliance.create_return_object()

def set(isamAppliance, name, users=None, groups=None, features=None, check_mode=False, force=False):
    """
    Configure a management authorization role
    """
    if _check(isamAppliance, name):
        # Update request
        return update(isamAppliance, name, users, groups, features, check_mode, force)
    else:
        # Force the add - we already know connection does not exist
        return add(isamAppliance, name, users, groups, features, check_mode, force)

def update(isamAppliance, name, users=None, groups=None, features=None, check_mode=False, force=False):
    """
    Update an existing management authorization role
    """
    _performUpdate = False
    if check_mode:
        return isamAppliance.create_return_object(changed=True)

    if force:
        _performUpdate = True
    else:
        """
        Compare
        """
        ret_obj = get(isamAppliance, name)
        newEntries = {
                      'name': name,
                      'users': users,
                      'groups': groups,
                      'features': features
                     }
        newEntriesJSON = json.dumps(newEntries, skipkeys=True, sort_keys=True)
        logger.debug(f"\nSorted management role configuration {name}:\n\n {newEntriesJSON}\n")
        currentEntriesJSON = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
        logger.debug(f"\nSorted Existing management role configuration {name}:\n\n {currentEntriesJSON}\n")
        if (newEntriesJSON != currentEntriesJSON):
            _performUpdate = True
    if not _performUpdate:
        logger.debug(f"No changes to role {name}")
        return isamAppliance.create_return_object(changed=False)
    else:
        return isamAppliance.invoke_put("Set management authorization role",
                                     f"/authorization/roles/{name}/v1", newEntries)

def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a management authorization role
    """
    if force is True or _check(isamAppliance, name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete management authorization role",
                f"/authorization/roles/{name}/v1")

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare all management authorization roles
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.isam.aac.server_connections.ws
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"



def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current external authentication service SCIM configuration
    """
    return isamAppliance.invoke_get("Retrieving the current external authentication service SCIM configuration ",
                                    f"{uri}/urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )


def set(isamAppliance, connection, schemas, check_mode=False, force=False):
    """
    Updating the external authentication service SCIM configuration settings
    """
    current_objs = get(isamAppliance)
    current_configs = current_objs['data']['urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS']
    ws_connections = ibmsecurity.isam.aac.server_connections.ws.get_all(isamAppliance)
    ws_connections = ws_connections['data']
    update_required = False
    uuid = None
    found = False

    for ws in ws_connections:
        if connection == ws['name']:
            uuid = ws['uuid']


    if uuid is None:
        warnings = f"Did not find connection {connection} in the configured server list."
        return isamAppliance.create_return_object(changed=False, warnings=warnings)

    for config in current_configs:
        if config['connection'] == uuid:
            found = True
            new_schemas = schemas
            old_schemas = config['schemas']
            sorted_new = json_sort(new_schemas)
            sorted_old = json_sort(old_schemas)
            if sorted_new != sorted_old:
                update_required = True

    if found is False:
        update_required = True


    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            data = []
            obj1 = {}
            obj1['connection'] = uuid
            obj1['schemas'] = schemas
            data.append(obj1)
            return isamAppliance.invoke_put("Updating the external authentication service SCIM configuration settings",
                                            f"{uri}/urn:ietf:params:scim:schemas:extension:isam:1.0:MMFA:EAS",
                                            data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version
                                            )

    return isamAppliance.create_return_object(changed=False)



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/fido2/metadata-services"
requires_modules = ["mga"]
requires_version = "10.0.4.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of FIDO2 Metadata Services
    """
    return isamAppliance.invoke_get("Retrieving the list of FIDO2 Metadata Services", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)

def get(isamAppliance, url, id=None, check_mode=False, force=False):
    """
    Retrieve a specific FIDO2 Metadata Service
    """
    if id is None:
        ret_obj = search(isamAppliance, url=url, check_mode=check_mode, force=force)
        id = ret_obj['data']

    if id == {}:
        logger.info(f"FIDO2 Metadata Service {url} had no match, skipping retrieval.")
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)

def search(isamAppliance, url, force=False, check_mode=False):
    """
    Search FIDO2 Metadata Service id by url
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['url'] == url:
            logger.info(f"Found FIDO2 Metadata Service {url} id: {obj['id']}")
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj

def delete(isamAppliance, url, check_mode=False, force=False):
    """
    Delete a FIDO2 Metadata Service
    """
    ret_obj = search(isamAppliance, url, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info(f"FIDO2 Metadata Service {url} not found, skipping delete.")
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a FIDO2 Metadata Service",
                f"{uri}/{id}",
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def set(isamAppliance, url, retryInterval=3600, truststore="", jwsTruststore="",
        username="", password="", keystore="", certificate="", protocol="", timeout=5, proxy="", headers=[],
        check_mode=False, force=False):
    """
    Create or Update a FIDO2 Metadata Service
    """
    if (search(isamAppliance, url=url))['data'] == {}:
        # Force the add - we already know FIDO2 Metadata Service does not exist
        logger.info(f"FIDO2 metadata service {url} had no match, requesting to add new one.")
        return add(isamAppliance, url=url, retryInterval=retryInterval, truststore=truststore, jwsTruststore=jwsTruststore,
                   username=username, password=password, keystore=keystore, certificate=certificate, protocol=protocol,
                   timeout=timeout, proxy=proxy, headers=headers, check_mode=check_mode, force=force)
    else:
        # Update request
        logger.info(f"FIDO2 metadata service {url} exists, requesting to update.")
        return update(isamAppliance, url=url, retryInterval=retryInterval, truststore=truststore, jwsTruststore=jwsTruststore,
                   username=username, password=password, keystore=keystore, certificate=certificate, protocol=protocol,
                   timeout=timeout, proxy=proxy, headers=headers, check_mode=check_mode, force=force)

def add(isamAppliance, url, retryInterval=3600, truststore="", jwsTruststore="",
        username="", password="", keystore="", certificate="", protocol="", timeout=5, proxy="", headers=[],
        check_mode=False, force=False):
    """
    Create a new FIDO2 Metadata Service
    """
    if force is False:
        ret_obj = search(isamAppliance, url)
    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new FIDO2 metadata service", uri,
                {
                    "url": url,
                    "retryInterval": int(retryInterval),
                    "truststore": truststore,
                    "jwsTruststore": jwsTruststore,
                    "username": username,
                    "password": password,
                    "keystore": keystore,
                    "certificate": certificate,
                    "protocol": protocol,
                    "timeout": int(timeout),
                    "proxy": proxy,
                    "headers": headers
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def update(isamAppliance, url="", retryInterval=3600, truststore="", jwsTruststore="",
           username="", password="", keystore="", certificate="", protocol="", timeout=5, proxy="", headers=[],
           check_mode=False, force=False):
    """
    Update a specific FIDO2 Metadata Service
    """
    retryInterval = int(retryInterval)
    timeout = int(timeout)
    mds_id, update_required, json_data = _check(isamAppliance, url, retryInterval, truststore, jwsTruststore,
                                                username, password, keystore, certificate, protocol,
                                                timeout, proxy, headers)
    if mds_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", f"Cannot update data for unknown FIDO2 relying party: {url}")

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific FIDO2 relying party",
                f"{uri}/{mds_id}", json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()

def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific FIDO2 Metadata Service by id",
                                    f"{uri}/{id}",
                                    requires_modules=requires_modules, requires_version=requires_version)

def _check(isamAppliance, url, retryInterval, truststore, jwsTruststore, username, password,
            keystore, certificate, protocol, timeout, proxy, headers):
    """
    Check if FIDO2 metadata service configuration is identical with server
    """
    update_required = False
    json_data = {}
    ret_obj = get(isamAppliance, url)
    if ret_obj['data'] == {}:
        logger.info("FIDO2 Metadata Service not found, returning update required.")
        return None, update_required, json_data
    else:
        logger.debug("Comparing server FIDO2 metadata service configuration with desired configuration.")

        # Converting python ret_obj['data'] dict to valid JSON (RFC 8259)
        # e.g. converts python boolean 'True' -> to JSON literal lowercase value 'true'
        cur_cfg = ret_obj['data']
        cur_cfg_id = cur_cfg['id']
        del cur_cfg['id']
        cur_json_string = json.dumps(cur_cfg)
        cur_sorted_json = tools.json_sort(cur_json_string)
        logger.debug(f"Server JSON : {cur_sorted_json}")

        new_cfg = {
            'url': url,
            'retryInterval': retryInterval,
            'truststore': truststore,
            'jwsTruststore': jwsTruststore,
            'username': username,
            'password': password,
            'keystore': keystore,
            'certificate': certificate,
            'protocol': protocol,
            'timeout': timeout,
            'proxy': proxy,
            'headers': headers
        }
        given_json_string = json.dumps(new_cfg)
        given_sorted_json = tools.json_sort(given_json_string)
        logger.debug(f"Desired JSON: {given_sorted_json}")

        if cur_sorted_json != given_sorted_json:
            logger.debug("Changes detected!")
            update_required = True
            json_data = new_cfg
        else:
            logger.debug("Server configuration is identical with desired configuration. No change necessary.")
            update_required = False
            json_data = new_cfg

        return cur_cfg_id, update_required, json_data



import json
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import logging
from ibmsecurity.appliance.ibmappliance import IBMAppliance
from ibmsecurity.appliance.isvgappliance import ISVGAppliance
from ibmsecurity.appliance.ibmappliance import IBMError
from ibmsecurity.utilities import tools

try:
    basestring
except NameError:
    basestring = (str, bytes)


class ISVGApplianceAdminProxy(ISVGAppliance):
    def __init__(self, adminProxyHostname, user, hostname, adminProxyProtocol='https', adminProxyPort=443,
                 adminProxyApplianceShortName=False):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating an ISVGAppliance over AdminProxy')

        self.adminProxyProtocol = adminProxyProtocol
        self.adminProxyHostname = adminProxyHostname

        # Type checking and tranformation to safely reuse this variable later on
        if isinstance(adminProxyPort, basestring):
            self.adminProxyPort = int(adminProxyPort)
        else:
            self.adminProxyPort = adminProxyPort

        self.adminProxyApplianceShortName = adminProxyApplianceShortName

        ISVGAppliance.__init__(self, hostname, user)

    def _url(self, uri):
        # shorten the junction name from hostname parameter
        # e.g.  isvg.ibm.com -junction-> /isvg          (with short name)
        #       isvg.ibm.com -junction-> /isvg.ibm.com  (without short name)
        if self.adminProxyApplianceShortName:
            applianceJunction = self.hostname.split('.')[0]
        else:
            applianceJunction = self.hostname

        # Build up the URL
        url = self.adminProxyProtocol + "://" + self.adminProxyHostname + ":" + str(
            self.adminProxyPort) + "/" + applianceJunction + uri
        self.logger.info("Issuing request to Appliance over AdminProxy: " + url)

        return url



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.appliance.ibmappliance import IBMError
from io import open

logger = logging.getLogger(__name__)
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    List the current nodes in the cluster
    """
    return isamAppliance.invoke_get("List the current nodes in the cluster",
                                    "/isam/cluster/nodes/v1", requires_model=requires_model)


def get_id(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the cluster identifier
    """
    return isamAppliance.invoke_get("Retrieve the cluster identifier",
                                    "/isam/cluster/id/v2", requires_model=requires_model)


def get_list(isamAppliance, check_mode=False, force=False):
    """
    Retrieving valid cluster identifiers
    """
    return isamAppliance.invoke_get("Retrieving valid cluster identifiers",
                                    "/isam/cluster/id/list/v2", requires_model=requires_model)


def get_default_id(isamAppliance, check_mode=False, force=False):
    """
    Retrieve default cluster identifier
    """
    return isamAppliance.invoke_get("Retrieve default cluster identifier",
                                    "/isam/cluster/id/default/v2", requires_model=requires_model)


def get_state(isamAppliance, cluster_id, check_mode=False, force=False):
    """
    Validate a cluster identifier
    """
    if not cluster_id:
        return isamAppliance.create_return_object()
    return isamAppliance.invoke_get("Validate a cluster identifier",
                                    f"/isam/cluster/id/address/{cluster_id}/state/v2",
                                    requires_model=requires_model)


def get_master(isamAppliance, check_mode=False, force=False):
    """
    Check if the node is the primary master
    """
    return isamAppliance.invoke_get("Check if the node is the primary master",
                                    "/isam/cluster/ismaster/v2", requires_model=requires_model)


def add_v2(isamAppliance, signature_file, cluster_id=None, restricted=False,
           check_mode=False, force=False):
    """
    Add a node to the cluster v2, accepting IP or hostname as cluster ID.
    If the cluster ID is not given, it falls back to the legacy behavior,
    to use the default address as the cluster ID.
    :param isamAppliance: The ISAM appliance to be added to the cluster.
    :param signature_file: The signature file to be used to join the cluster.
    :param cluster_id: The cluster ID to be used by the target node.
                        Could be a hostname or an IP address.
    :param restricted: Whether it is a restricted node.
    :return: True if succeeded.
    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post_files(
            "Add a node to the cluster v2",
            "/isam/cluster/nodes/v2",
            [
                {
                    'file_formfield': 'signature_file',
                    'filename': signature_file,
                    'mimetype': 'application/octet-stream'
                }
            ],
            {'restricted': restricted,
             'signature_file': open(signature_file, 'rb'),
             'address': cluster_id,
             }, json_response=False, requires_model=requires_model)


def add(isamAppliance, signature_file, restricted=False, check_mode=False, force=False):
    """
    Add a node to the cluster
    """

    id = None
    if force is False:
        try:
            ret_obj = get_id(isamAppliance)

            if 'Docker' in ' '.join(ret_obj['warnings']):
                return isamAppliance.create_return_object(warnings=ret_obj['warnings'])
            else:
                id = ret_obj['data']['value']
        except IBMError:
            ret_obj = get_default_id(isamAppliance)

            if 'Docker' in ' '.join(ret_obj['warnings']):
                return isamAppliance.create_return_object(warnings=ret_obj['warnings'])
            else:
                id = ret_obj['data']['address']

    check_obj = _check(isamAppliance, id)
    if force is True or check_obj['value'] is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=check_obj['warnings'])
        else:
            return isamAppliance.invoke_post_files(
                "Add a node to the cluster",
                "/isam/cluster/nodes/v1",
                [
                    {
                        'file_formfield': 'signature_file',
                        'filename': signature_file,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {'restricted': restricted}, json_response=False, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=check_obj['warnings'])


def _check(isamAppliance, id):
    obj = {'value': False, 'warnings': ""}

    ret_obj = get_all(isamAppliance)
    obj['warnings'] = ret_obj['warnings']
    for node in ret_obj['data']:
        if node['address'] == id:
            obj['value'] = True
            return obj

    obj['value'] = False
    return obj


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Remove a node from the cluster
    """

    check_obj = _check(isamAppliance, id)
    if force is True or check_obj['value'] is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=check_obj['warnings'])
        else:
            return isamAppliance.invoke_delete(
                "Remove a node from the cluster",
                f"/isam/cluster/nodes/{id}/v1", requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=check_obj['warnings'])


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cluster nodes between two appliances
    """

    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import json
import tempfile
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/custfile_mgmt"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Get nls property file info
    """
    return isvgAppliance.invoke_get("Retrieving nls file entries", uri + "/directories")


def search(isvgAppliance, filename, check_mode=False, force=False):
    """
    Search for existing nls property by file name
    """
    ret_obj = get_all(isvgAppliance)

    return_obj = isvgAppliance.create_return_object()

    if 'children' in ret_obj['data']:
        for child in ret_obj['data']['children']:
            if 'fileName' in child and child['fileName'] == 'nls':
                files=json.loads(child['files'])
                for file in files:
                    if 'fileName' in file and file['fileName'] == filename:
                        logger.info("Found filename entry: {0}".format(file['fileName']))
                        return_obj['data'] = file
                        return_obj['rc'] = 0
                        break

    logger.debug("Found nls entry: {0}".format(return_obj))

    return return_obj


def upload(isvgAppliance, filename, check_mode=False, force=False):
    """
    Import nls property file
    """
    warnings = []

    if force is True or not _check(isvgAppliance, filename):
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            file_short = tools.path_leaf(filename)
            ret_obj = isvgAppliance.invoke_post_files(
                "Import nls property",
                "/upload_object",
                [
                    {
                        'file_formfield': 'uploadedfile',
                        'filename': filename,
                        'mimetype': 'text/plain'
                    }
                ],
                {
                    'fileName': file_short,
                    'upload_file_type': 'custfile'
                },
                json_response=False)

            warnings = ret_obj['warnings']

            if ret_obj["rc"] == 0:
                nls_json = {
                   "fileName": file_short,
                   "filePath": "nls/",
                   "ObjType": "externalfile"
                }

                return isvgAppliance.invoke_post(
                    "Save nls property metadata", uri, nls_json, warnings=warnings)

    return isvgAppliance.create_return_object()


def download(isvgAppliance, filename, check_mode=False, force=False):
    """
    Download nls property file
    """
    import os.path

    if force is True or (os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isvgAppliance.invoke_get_file("Downloading nls file",
                                                 "{0}/{1}?nls/{2}".format(uri, "download", tools.path_leaf(filename)), filename=filename, mime_types="text/html")

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, filename):
    """
    Check if nls filename needs to be uploaded
    Need to download current filename in order to compare both
    """
    ret_obj = search(isvgAppliance, tools.path_leaf(filename))

    if len(ret_obj['data']) > 0:
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpFile = tmpdirname + "/" + tools.path_leaf(filename)
            ret_obj = download(isvgAppliance, tmpFile)
            if ret_obj["rc"] == 0:
                return tools.files_same(filename, tmpFile)
    else:
        return True



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/configuration/logging"
requires_modules = None
requires_versions = None
requires_model="Appliance"


def get(isamAppliance):
    """
    Retrieving logging configuration attributes
    """
    return isamAppliance.invoke_get("Retrieving logging configuration attributes", module_uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_versions,
                                    requires_model=requires_model)


def update(isamAppliance, local, remote_address, remote_port, remote_facility, check_mode=False, force=False):
    """
    Updates logging configuration
    """
    json_data = {
        "local": local,
        "remote_address": remote_address,
        "remote_port": remote_port,
        "remote_facility": remote_facility
    }

    change_required, warnings = _check(isamAppliance, json_data)

    if force is True or change_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating logging configuration attributes", module_uri, json_data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_versions,
                                            requires_model=requires_model)

    else:
        return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, json_data):
    """
    Checks update for idempotency
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']
    change_required = False

    if json_data['local'] is True:
        if 'local' in ret_obj['data']:
            if json_data['local'] != ret_obj['data']['local']:
                change_required = True
    else:
        sorted_ret_obj = tools.json_sort(ret_obj['data'])
        sorted_json_data = tools.json_sort(json_data)

        logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
        logger.debug(f"Sorted Desired  Data:{sorted_json_data}")

        if sorted_ret_obj != sorted_json_data:
            change_required = True

    return change_required, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare configuration between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/isam/container_ext/repo"

requires_version = "10.0.7.0"
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving health for all (running?) containers
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.repo.get_all

    """
    return isamAppliance.invoke_get(
        "Retrieving all Registry Authentication",
        f"{uri}",
        requires_model=requires_model,
        requires_version=requires_version,
    )


def get(isamAppliance, registry_id, check_mode=False, force=False):
    """
    Get credential for known users of container registry
    """
    return isamAppliance.invoke_get(
        "Retrieving health for container",
        f"{uri}/{registry_id}",
        requires_model=requires_model,
        requires_version=requires_version,
    )


def add(
    isamAppliance,
    registry,
    user=None,
    secret=None,
    **kwargs,
):
    """
    Add a credential for a user and container registry

    :param isamAppliance:
    :param registry: hostname of the registry, eg. icr.io
    :param user: user (optional)
    :param secret: the secret for the user (optional)
    :param proxy_host (optional)
    :param proxy_port (optional)
    :param proxy_user (optional)
    :param proxy_pass (optional)
    :param proxy_schema (optional)
    :param warnings
    :param check_mode
    :param force
    """
    input_args = {}
    force = False
    check_mode = False
    warnings = []
    for k, v in kwargs.items():
        if k == "force":
            force = v
            continue
        if k == "check_mode":
            check_mode = v
            continue
        if k == "warnings":
            warnings = v
        input_args[k] = v

    if force or not _check(isamAppliance, registry, user):
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            post_data = {
                "host": registry,
                "user": user,
                "secret": secret,
            }

            post_data.update(input_args)

            return isamAppliance.invoke_post(
                "Create new registry",
                uri,
                post_data,
                requires_model=requires_model,
                requires_version=requires_version,
                warnings=warnings,
            )

    return isamAppliance.create_return_object()


def update(
    isamAppliance,
    registry,
    user=None,
    secret=None,
    **kwargs,
):
    """
    Update a credential for a user and container registry
    (user and secret are NOT required values, although it is stated like that in the documentation)

    :param isamAppliance:
    :param registry: hostname of the registry, eg. icr.io
    :param user: user
    :param secret: the secret for the user
    :param warnings
    :param check_mode
    :param force
    """
    input_args = {}
    force = False
    check_mode = False
    warnings = []
    for k, v in kwargs.items():
        if k == "force":
            force = v
            continue
        if k == "check_mode":
            check_mode = v
            continue
        if k == "warnings":
            warnings = v
        input_args[k] = v
    registry_id = search(isamAppliance, registry, user)
    registry_id = registry_id.get('data', None)
    if force or registry_id is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if tools.version_compare(isamAppliance.facts['version'], "10.0.9.0") < 0:
                warnings.append(f"Appliance is at version: {isamAppliance.facts['version']}. Using the previous format")
                put_data = {
                    "host": registry,
                    "user": user,
                }

            else:
                put_data = {
                    "host": registry,
                    "user": user,
                }
                # append other fields from input parameters
                put_data.update(input_args)
            if secret is not None:
                put_data['secret'] = secret

            return isamAppliance.invoke_put(
                "Update registry",
                f"{uri}/{registry_id}",
                put_data,
                requires_model=requires_model,
                requires_version=requires_version,
                warnings=warnings,
            )

    return isamAppliance.create_return_object()


def set(
    isamAppliance,
    registry,
    user=None,
    secret=None,
    **kwargs,
):
    """
    Set - we can't make this function idempotent if a secret is set.
    (user and secret are NOT required values, although it is stated like that in the documentation)
    TODO: execute every time ?
    """
    if _check(isamAppliance, registry, user):
        logger.debug(f"\nUpdating {registry} with {user}")
        # TODO: make this idempotent
        return update(
            isamAppliance,
            registry,
            user,
            secret,
            warnings=["No idempotency because of secrets"],
            **kwargs)
    else:
        logger.debug(f"\nCreating {registry} with {user}")
        return add(
            isamAppliance,
            registry,
            user,
            secret,
            **kwargs)


def search(isamAppliance, host, user=None, check_mode=False, force=False):
    """
    Return the id of the repository

    :param isamAppliance:
    :param host:
    :param user:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = get_all(isamAppliance, check_mode, force)
    return_obj = isamAppliance.create_return_object()
    return_obj["data"] = None
    for obj in ret_obj["data"]:
        if user is None:
            if obj.get("host", "") == host:
                return_obj["data"] = obj["id"]
                return_obj["rc"] = 0
                break
        else:
            if obj.get("host", "") == host and obj.get("user", "") == user:
                return_obj["data"] = obj["id"]
                return_obj["rc"] = 0
                break

    return return_obj


def _check(isamAppliance, registry, user):
    """
    Check if there's a repository already for this host
    I'll make a combination of registry and user....

    """
    ret_obj = get_all(isamAppliance)
    for c in ret_obj["data"]:
        if c.get("host") == registry and c.get("user") == user:
            logger.debug(f"Repository by host {registry} and user {user} exists...")
            return True
            break
    return False


def delete(isamAppliance, registry, user=None, check_mode=False, force=False):
    """
    Delete a repo by host (and user)
    """
    ret_obj = search(isamAppliance, registry, user)
    registry_id = ret_obj.get("data", None)

    if force or registry_id is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a repository configuration",
                f"{uri}/{registry_id}"
            )

    return isamAppliance.create_return_object()



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration/schema_extensions"
requires_modules = ["mga"]
requires_version = "10.0.4"

"""
Manage SCIM Custom Schema Extensions

The array of attributes must be included in create and update requests as detailed below.

Attribute parameters include:
name
description
type - string, boolean, decimal, integer, dateTime, reference, complex (object), binary
mapping.sourceType - ldap, session, fixed
mapping.sourceAttribute
multiValued - True/False
required - True/False
mutability - readOnly, readWrite, immutable, writeOnly, adminWrite, userWrite
returned - always, never, default, request
canonicalValues
subAttributes

For example:

[
    {
        "name": "memberNumber",
        "description": "",
        "type": "integer",
        "mapping": {
            "sourceType": "ldap",
            "sourceAttribute": "membershipNumber"
        },
        "multiValued": False,
        "required": False,
        "mutability": "immutable",
        "returned": "default",
        "canonicalValues": [1, 2, 3]
    },
    {
        "name": "department",
        "description": "",
        "mutability": "readWrite",
        "type": "complex",
        "multiValued": False,
        "returned": "default",
        "required": False,
        "subAttributes": [
            {
                "mapping": {
                "sourceType": "ldap",
                "sourceAttribute": "department"
                },
                "name": "name",
                "description": "",
                "mutability": "readWrite",
                "type": "string",
                "multiValued": False,
                "returned": "default",
                "required": False
            },
            {
                "mapping": {
                "sourceType": "ldap",
                "sourceAttribute": "departmentCode"
                },
                "name": "code",
                "description": "",
                "mutability": "readWrite",
                "type": "string",
                "multiValued": False,
                "returned": "default",
                "required": False
            }
        ]
    }
]

"""

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the complete list of SCIM Custom Schema Extensions
    """
    return isamAppliance.invoke_get("Retrieving the complete list of SCIM Custom Schema Extensions",
                                    uri)


def get(isamAppliance, urn, check_mode=False, force=False):
    """
    Retrieving the configuration of a SCIM Custom Schema Extension by urn/id
    """
    return isamAppliance.invoke_get("Retrieve the configuration of a SCIM Custom Schema Extension by urn/id",
                                       f"{uri}/{urn}")


def set(isamAppliance, urn, name, description='', attributes=None,
        new_name=None, check_mode=False, force=False):
    """
    Creating or Modifying a SCIM Custom Schema Extension
    """
    if _check_exists(isamAppliance, urn=urn, name=name) is False:
        # Force the add - we already know connection does not exist
        return add(isamAppliance=isamAppliance, urn=urn, name=name, description=description,
                    attributes=attributes, check_mode=check_mode, force=True)
    else:
        # Update request
        return update(isamAppliance=isamAppliance, urn=urn, name=name, description=description,
                      attributes=attributes, new_name=new_name, check_mode=check_mode, force=force)


def add(isamAppliance, urn, name, description='', attributes=None, check_mode=False, force=False):
    """
    Creating a SCIM Custom Schema Extension
    """
    if force is True or _check_exists(isamAppliance, urn=urn, name=name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a SCIM Custom Schema Extension",
                uri,
                _create_json(urn=urn, name=name, description=description, attributes=attributes),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, urn, name, description='', attributes=None, new_name=None, check_mode=False, force=False):
    """
    Modifying a SCIM Custom Schema Extension

    Use new_name to rename the extension.
    """
    ret_obj = get(isamAppliance, urn)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append(f"SCIM Custom Schema Extension {urn} not found, skipping update.")
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        id = ret_obj["data"]["id"]

    needs_update = False

    json_data = _create_json(urn=urn, name=name, description=description, attributes=attributes)
    if new_name is not None:  # Rename condition
        json_data['name'] = new_name

    if force is not True:
        sorted_ret_obj = tools.json_sort(ret_obj['data'])
        sorted_json_data = tools.json_sort(json_data)
        logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
        logger.debug(f"Sorted Desired  Data:{sorted_json_data}")

        if sorted_ret_obj != sorted_json_data:
            needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Modifying a SCIM Custom Schema Extension",
               f"{uri}/{urn}", json_data,
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, urn=None, name=None, check_mode=False, force=False):
    """
    Deleting a SCIM Custom Schema Extension
    """
    if force is True or _check_exists(isamAppliance, urn=urn, name=name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if urn == None and name != None:
                ret_obj = search(isamAppliance, name=name)
                urn = ret_obj['data']

            return isamAppliance.invoke_delete(
                "Deleting a SCIM Custom Schema Extension",
                f"{uri}/{urn}",
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def search(isamAppliance, name):
    """
    Retrieve URN for named SCIM Custom Schema Extension
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            ret_obj_new['data'] = obj['id']

    return ret_obj_new


def _create_json(urn, name, description, attributes):
    """
    Create a JSON to be used for the REST API call
    """

    json = {
        "id": urn,
        "name": name,
        "description": description,
        "attributes": attributes
    }

    return json


def _check_exists(isamAppliance, urn, name):
    """
    Check if the SCIM Custom Schema Extension already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if (name is not None and obj['name'] == name) or (urn is not None and obj['id'] == urn):
            return True

    return False
